#!/usr/bin/env python

import getpass
import os

SECRET_FILE = "secret.txt"


def add_credentials():
    with open(SECRET_FILE, "a") as f:
        f.write(str(input("Enter your username: ")) + ":")
        print(
            "Enter your password: (it will not be shown due to security "
            + "reasons - just start typing and press Enter)"
        )
        f.write(getpass.getpass() + "\n")


def get_credentials(username=None):
    """Returns login and password stored in `secret.txt`."""
    while not check_secret():
        pass
    while True:
        try:
            with open(SECRET_FILE, "r") as f:
                lines = [line.strip().split(":", 2) for line in f.readlines()]
        except ValueError:
            msg = "Problem with opening `{}`, will remove the file."
            raise Exception(msg.format(SECRET_FILE))
        if username is not None:
            for login, password in lines:
                if login == username.strip():
                    return login, password
        print("Which account do you want to use? (Type number)")
        for ind, (login, password) in enumerate(lines):
            print("%d: %s" % (ind + 1, login))
        print("%d: %s" % (0, "add another account."))
        print("%d: %s" % (-1, "delete all accounts."))
        try:
            ind = int(input())
            if ind == 0:
                add_credentials()
                continue
            elif ind == -1:
                delete_credentials()
                check_secret()
                continue
            elif 0 <= ind - 1 < len(lines):
                return lines[ind - 1]
        except Exception:
            print("Wrong input, enter the number of the account to use.")


def check_secret():
    while True:
        if os.path.exists(SECRET_FILE):
            with open(SECRET_FILE, "r") as f:
                try:
                    login, password = f.readline().strip().split(":")
                    if len(login) < 4 or len(password) < 6:

                        print(
                            "Data in `secret.txt` file is invalid. "
                            "We will delete it and try again."
                        )

                        os.remove(SECRET_FILE)
                    else:
                        return True
                except Exception:
                    print("Your file is broken. We will delete it " + "and try again.")
                    os.remove(SECRET_FILE)
        else:
            print(
                "We need to create a text file '%s' where "
                "we will store your login and password from Instagram." % SECRET_FILE
            )
            print("Don't worry. It will be stored locally.")
            while True:
                add_credentials()
                print("Do you want to add another account? (y/n)")
                if input().lower() != "y":
                    break


def delete_credentials():
    if os.path.exists(SECRET_FILE):
        os.remove(SECRET_FILE)


if __name__ == "__main__":
    check_secret()



from tqdm import tqdm


def delete_media(self, media_id):
    self.small_delay()
    media = self.get_media_info(media_id)
    media = media[0] if isinstance(media, list) else media
    if self.api.delete_media(media):
        return True
    self.logger.info("Media with {} is not {}.".format(media.get("id"), "deleted"))
    return False


def delete_medias(self, medias):
    broken_items = []
    if not medias:
        self.logger.info("Nothing to delete.")
        return broken_items
    self.logger.info("Going to delete %d medias." % (len(medias)))
    for media in tqdm(medias):
        if not self.delete_media(media):
            self.error_delay()
            broken_items = medias[medias.index(media) :]
            break
    self.logger.info(
        "DONE: Total deleted %d medias." % (len(medias) - len(broken_items))
    )
    return broken_items


def delete_comment(self, media_id, comment_id):
    if self.api.delete_comment(media_id, comment_id):
        self.small_delay()
        return True
    self.logger.info(
        "Comment with {} in media {} is not deleted.".format(comment_id, media_id)
    )
    return False



# flake8: noqa
# Devices taken from:
# https://github.com/mgp25/Instagram-API/blob/
#   master/src/Devices/GoodDevices.php
DEFAULT_DEVICE = "one_plus_7"
# from https://www.apkmirror.com/apk/instagram/instagram-instagram/
#   instagram-instagram-105-0-0-18-119-release/instagram-105-0-0-18
#   -119-4-android-apk-download/
INSTAGRAM_VERSION = "105.0.0.18.119"
DEVICES = {
    # Released on August 2019
    "one_plus_7": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 28,
        "android_release": "9.0",
        "dpi": "420dpi",
        "resolution": "1080x2260",
        "manufacturer": "OnePlus",
        "device": "GM1903",
        "model": "OnePlus7",
        "cpu": "qcom",
    },
    # Released on March 2016
    "samsung_galaxy_s7": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 24,
        "android_release": "7.0",
        "dpi": "640dpi",
        "resolution": "1440x2560",
        "manufacturer": "samsung",
        "device": "SM-G930F",
        "model": "herolte",
        "cpu": "samsungexynos8890",
    },
    # Released on January 2017
    "huawei_mate_9_pro": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 24,
        "android_release": "7.0",
        "dpi": "640dpi",
        "resolution": "1440x2560",
        "manufacturer": "HUAWEI",
        "device": "LON-L29",
        "model": "HWLON",
        "cpu": "hi3660",
    },
    # Released on February 2018
    "samsung_galaxy_s9_plus": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 24,
        "android_release": "7.0",
        "dpi": "640dpi",
        "resolution": "1440x2560",
        "manufacturer": "samsung",
        "device": "SM-G965F",
        "model": "star2qltecs",
        "cpu": "samsungexynos9810",
    },
    # Released on November 2016
    "one_plus_3t": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 24,
        "android_release": "7.0",
        "dpi": "380dpi",
        "resolution": "1080x1920",
        "manufacturer": "OnePlus",
        "device": "ONEPLUS A3010",
        "model": "OnePlus3T",
        "cpu": "qcom",
    },
    # Released on April 2016
    "lg_g5": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 23,
        "android_release": "6.0.1",
        "dpi": "640dpi",
        "resolution": "1440x2392",
        "manufacturer": "LGE/lge",
        "device": "RS988",
        "model": "h1",
        "cpu": "h1",
    },
    # Released on June 2016
    "zte_axon_7": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 23,
        "android_release": "6.0.1",
        "dpi": "640dpi",
        "resolution": "1440x2560",
        "manufacturer": "ZTE",
        "device": "ZTE A2017U",
        "model": "ailsa_ii",
        "cpu": "qcom",
    },
    # Released on March 2016
    "samsung_galaxy_s7_edge": {
        "instagram_version": INSTAGRAM_VERSION,
        "android_version": 23,
        "android_release": "6.0.1",
        "dpi": "640dpi",
        "resolution": "1440x2560",
        "manufacturer": "samsung",
        "device": "SM-G935",
        "model": "hero2lte",
        "cpu": "samsungexynos8890",
    },
}



import pytest
import responses

from instabot.api.config import API_URL, SIG_KEY_VERSION

from .test_bot import TestBot
from .test_variables import (
    TEST_FOLLOWER_ITEM,
    TEST_FOLLOWING_ITEM,
    TEST_SEARCH_USERNAME_ITEM,
    TEST_USERNAME_INFO_ITEM,
)

try:
    from unittest.mock import patch
except ImportError:
    from mock import patch


def reset_files(_bot):
    for x in _bot.followed_file.list:
        _bot.followed_file.remove(x)
    for x in _bot.unfollowed_file.list:
        _bot.unfollowed_file.remove(x)
    for x in _bot.skipped_file.list:
        _bot.skipped_file.remove(x)


class TestBotFilter(TestBot):
    @responses.activate
    @pytest.mark.parametrize(
        "username",
        [
            TEST_SEARCH_USERNAME_ITEM["username"],
            TEST_SEARCH_USERNAME_ITEM["pk"],
            str(TEST_SEARCH_USERNAME_ITEM["pk"]),
        ],
    )
    @patch("time.sleep", return_value=None)
    def test_follow(self, patched_time_sleep, username):
        follows_at_start = self.bot.total["follows"]
        self.bot._following = [1]
        reset_files(self.bot)
        user_id = TEST_SEARCH_USERNAME_ITEM["pk"]
        my_test_search_username_item = TEST_SEARCH_USERNAME_ITEM.copy()
        my_test_username_info_item = TEST_USERNAME_INFO_ITEM.copy()
        my_test_search_username_item["is_verified"] = False
        my_test_search_username_item["is_business"] = False
        my_test_search_username_item["is_private"] = False
        my_test_search_username_item["follower_count"] = 100
        my_test_search_username_item["following_count"] = 15
        my_test_search_username_item["media_count"] = (
            self.bot.min_media_count_to_follow + 1
        )
        my_test_search_username_item["has_anonymous_profile_picture"] = False
        my_test_username_info_item["pk"] = TEST_SEARCH_USERNAME_ITEM["pk"]
        my_test_username_info_item["username"] = TEST_SEARCH_USERNAME_ITEM["username"]
        my_test_username_info_item["is_verified"] = False
        my_test_username_info_item["is_business"] = False
        my_test_username_info_item["is_private"] = False
        my_test_username_info_item["follower_count"] = 100
        my_test_username_info_item["following_count"] = 15
        my_test_username_info_item["media_count"] = (
            self.bot.min_media_count_to_follow + 1
        )
        my_test_username_info_item["has_anonymous_profile_picture"] = False

        response_data = {"status": "ok", "user": my_test_search_username_item}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=username
            ),
            status=200,
            json=response_data,
        )

        response_data = {"status": "ok", "user": my_test_username_info_item}
        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(api_url=API_URL, user_id=user_id),
            status=200,
            json=response_data,
        )

        response_data = {"status": "ok"}
        responses.add(
            responses.POST,
            "{api_url}friendships/create/{user_id}/".format(
                api_url=API_URL, user_id=user_id
            ),
            json=response_data,
            status=200,
        )

        assert self.bot.follow(username)
        assert self.bot.total["follows"] == follows_at_start + 1
        assert self.bot.followed_file.list[-1] == str(user_id)
        assert str(user_id) in self.bot.following

    @responses.activate
    @pytest.mark.parametrize(
        "user_ids",
        [
            [
                str(TEST_SEARCH_USERNAME_ITEM["pk"]),
                str(TEST_SEARCH_USERNAME_ITEM["pk"] + 1),
                str(TEST_SEARCH_USERNAME_ITEM["pk"] + 2),
                str(TEST_SEARCH_USERNAME_ITEM["pk"] + 3),
            ],
            [
                str(TEST_SEARCH_USERNAME_ITEM["pk"]),
                str(TEST_SEARCH_USERNAME_ITEM["pk"] + 4),
                str(TEST_SEARCH_USERNAME_ITEM["pk"] + 5),
                str(TEST_SEARCH_USERNAME_ITEM["pk"] + 6),
            ],
        ],
    )
    @patch("time.sleep", return_value=None)
    def test_follow_users(self, patched_time_sleep, user_ids):
        self.bot._following = [1]
        reset_files(self.bot)
        follows_at_start = self.bot.total["follows"]
        self.bot.followed_file.append(str(user_ids[1]))
        self.bot.unfollowed_file.append(str(user_ids[2]))
        self.bot.skipped_file.append(str(user_ids[3]))

        my_test_search_username_item = TEST_SEARCH_USERNAME_ITEM.copy()
        my_test_username_info_item = TEST_USERNAME_INFO_ITEM.copy()

        my_test_search_username_item["is_verified"] = False
        my_test_search_username_item["is_business"] = False
        my_test_search_username_item["is_private"] = False
        my_test_search_username_item["follower_count"] = 100
        my_test_search_username_item["following_count"] = 15
        my_test_search_username_item["media_count"] = (
            self.bot.min_media_count_to_follow + 1
        )
        my_test_search_username_item["has_anonymous_profile_picture"] = False
        my_test_username_info_item["username"] = TEST_SEARCH_USERNAME_ITEM["username"]
        my_test_username_info_item["is_verified"] = False
        my_test_username_info_item["is_business"] = False
        my_test_username_info_item["is_private"] = False
        my_test_username_info_item["follower_count"] = 100
        my_test_username_info_item["following_count"] = 15
        my_test_username_info_item["media_count"] = (
            self.bot.min_media_count_to_follow + 1
        )
        my_test_username_info_item["has_anonymous_profile_picture"] = False

        for user_id in user_ids:
            my_test_search_username_item["pk"] = user_id
            my_test_username_info_item["pk"] = user_id

            response_data = {"status": "ok", "user": my_test_search_username_item}
            responses.add(
                responses.GET,
                "{api_url}users/{username}/usernameinfo/".format(
                    api_url=API_URL, username=user_id
                ),
                status=200,
                json=response_data,
            )

            response_data = {"status": "ok", "user": my_test_username_info_item}
            responses.add(
                responses.GET,
                "{api_url}users/{user_id}/info/".format(
                    api_url=API_URL, user_id=user_id
                ),
                status=200,
                json=response_data,
            )

            response_data = {"status": "ok"}
            responses.add(
                responses.POST,
                "{api_url}friendships/create/{user_id}/".format(
                    api_url=API_URL, user_id=user_id
                ),
                json=response_data,
                status=200,
            )

        test_broken_items = [] == self.bot.follow_users(user_ids)
        test_follows = self.bot.total["follows"] == follows_at_start + 1
        test_following = self.bot.following == [1, user_ids[0]]
        test_followed = str(user_ids[0]) in self.bot.followed_file.list
        assert test_broken_items and test_follows and test_followed and test_following

    @responses.activate
    @pytest.mark.parametrize("username", ["1234567890", 1234567890])
    @patch("time.sleep", return_value=None)
    def test_follow_followers(self, patched_time_sleep, username):
        self.blacklist = []
        my_test_search_username_item = TEST_SEARCH_USERNAME_ITEM.copy()
        my_test_username_info_item = TEST_USERNAME_INFO_ITEM.copy()
        my_test_follower_item = TEST_FOLLOWER_ITEM.copy()

        self.bot._following = []
        reset_files(self.bot)
        follows_at_start = self.bot.total["follows"]

        response_data_1 = {"status": "ok", "user": TEST_SEARCH_USERNAME_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=username
            ),
            status=200,
            json=response_data_1,
        )

        response_data_2 = {"status": "ok", "user": TEST_USERNAME_INFO_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(api_url=API_URL, user_id=username),
            status=200,
            json=response_data_2,
        )

        results_3 = 5
        my_test_follower_items = [
            my_test_follower_item.copy() for _ in range(results_3)
        ]
        my_test_search_username_items = [
            my_test_search_username_item.copy() for _ in range(results_3)
        ]
        my_test_username_info_items = [
            my_test_username_info_item.copy() for _ in range(results_3)
        ]

        for i, _ in enumerate(range(results_3)):
            my_test_follower_items[i]["pk"] = TEST_FOLLOWER_ITEM["pk"] + i
            my_test_follower_items[i]["username"] = "{}_{}".format(
                TEST_FOLLOWER_ITEM["username"], i
            )
        response_data_3 = {
            "status": "ok",
            "big_list": False,
            "next_max_id": None,
            "sections": None,
            "users": my_test_follower_items,
        }
        responses.add(
            responses.GET,
            (
                "{api_url}friendships/{user_id}/followers/?" + "rank_token={rank_token}"
            ).format(
                api_url=API_URL, user_id=username, rank_token=self.bot.api.rank_token
            ),
            json=response_data_3,
            status=200,
        )

        for i, _ in enumerate(range(results_3)):
            my_test_search_username_items[i]["username"] = "{}_{}".format(
                TEST_FOLLOWER_ITEM["username"], i
            )
            my_test_search_username_items[i]["pk"] = TEST_FOLLOWER_ITEM["pk"] + i
            my_test_search_username_items[i]["is_verified"] = False
            my_test_search_username_items[i]["is_business"] = False
            my_test_search_username_items[i]["is_private"] = False
            my_test_search_username_items[i]["follower_count"] = 100
            my_test_search_username_items[i]["following_count"] = 15
            my_test_search_username_items[i]["media_count"] = (
                self.bot.min_media_count_to_follow + 1
            )
            my_test_search_username_items[i]["has_anonymous_profile_picture"] = False

            my_test_username_info_items[i]["username"] = "{}_{}".format(
                TEST_FOLLOWER_ITEM["username"], i
            )
            my_test_username_info_items[i]["pk"] = TEST_FOLLOWER_ITEM["pk"] + i
            my_test_username_info_items[i]["is_verified"] = False
            my_test_username_info_items[i]["is_business"] = False
            my_test_username_info_items[i]["is_private"] = False
            my_test_username_info_items[i]["follower_count"] = 100
            my_test_username_info_items[i]["following_count"] = 15
            my_test_username_info_items[i]["media_count"] = (
                self.bot.min_media_count_to_follow + 1
            )
            my_test_username_info_items[i]["has_anonymous_profile_picture"] = False

            response_data = {"status": "ok", "user": my_test_search_username_items[i]}
            responses.add(
                responses.GET,
                "{api_url}users/{username}/usernameinfo/".format(
                    api_url=API_URL,
                    username=my_test_search_username_items[i]["username"],
                ),
                status=200,
                json=response_data,
            )

            response_data = {"status": "ok", "user": my_test_username_info_items[i]}
            responses.add(
                responses.GET,
                "{api_url}users/{user_id}/info/".format(
                    api_url=API_URL, user_id=my_test_username_info_items[i]["pk"]
                ),
                status=200,
                json=response_data,
            )

            response_data = {"status": "ok"}
            responses.add(
                responses.POST,
                "{api_url}friendships/create/{user_id}/".format(
                    api_url=API_URL, user_id=my_test_username_info_items[i]["pk"]
                ),
                json=response_data,
                status=200,
            )

        self.bot.follow_followers(username)

        test_follows = self.bot.total["follows"] == follows_at_start + results_3
        test_following = sorted(self.bot.following) == [
            str(my_test_username_info_items[i]["pk"]) for i in range(results_3)
        ]
        test_followed = sorted(self.bot.followed_file.list) == [
            str(my_test_username_info_items[i]["pk"]) for i in range(results_3)
        ]
        assert test_follows and test_following and test_followed

    @responses.activate
    @pytest.mark.parametrize("username", ["1234567890", 1234567890])
    @patch("time.sleep", return_value=None)
    def test_follow_following(self, patched_time_sleep, username):
        self.blacklist = []
        my_test_search_username_item = TEST_SEARCH_USERNAME_ITEM.copy()
        my_test_username_info_item = TEST_USERNAME_INFO_ITEM.copy()
        my_test_following_item = TEST_FOLLOWING_ITEM.copy()

        self.bot._following = []
        reset_files(self.bot)
        follows_at_start = self.bot.total["follows"]

        response_data_1 = {"status": "ok", "user": TEST_SEARCH_USERNAME_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{username}/usernameinfo/".format(
                api_url=API_URL, username=username
            ),
            status=200,
            json=response_data_1,
        )

        response_data_2 = {"status": "ok", "user": TEST_USERNAME_INFO_ITEM}
        responses.add(
            responses.GET,
            "{api_url}users/{user_id}/info/".format(api_url=API_URL, user_id=username),
            status=200,
            json=response_data_2,
        )

        results_3 = 5
        my_test_following_items = [
            my_test_following_item.copy() for _ in range(results_3)
        ]
        my_test_search_username_items = [
            my_test_search_username_item.copy() for _ in range(results_3)
        ]
        my_test_username_info_items = [
            my_test_username_info_item.copy() for _ in range(results_3)
        ]

        for i, _ in enumerate(range(results_3)):
            my_test_following_items[i]["pk"] = TEST_FOLLOWING_ITEM["pk"] + i
            my_test_following_items[i]["username"] = "{}_{}".format(
                TEST_FOLLOWING_ITEM["username"], i
            )
        response_data_3 = {
            "status": "ok",
            "big_list": False,
            "next_max_id": None,
            "sections": None,
            "users": my_test_following_items,
        }
        responses.add(
            responses.GET,
            (
                "{api_url}friendships/{user_id}/following/?max_id={max_id}&"
                + "ig_sig_key_version={sig_key}&rank_token={rank_token}"
            ).format(
                api_url=API_URL,
                user_id=username,
                rank_token=self.bot.api.rank_token,
                sig_key=SIG_KEY_VERSION,
                max_id="",
            ),
            json=response_data_3,
            status=200,
        )

        for i, _ in enumerate(range(results_3)):
            my_test_search_username_items[i]["username"] = "{}_{}".format(
                TEST_FOLLOWING_ITEM["username"], i
            )
            my_test_search_username_items[i]["pk"] = TEST_FOLLOWING_ITEM["pk"] + i
            my_test_search_username_items[i]["is_verified"] = False
            my_test_search_username_items[i]["is_business"] = False
            my_test_search_username_items[i]["is_private"] = False
            my_test_search_username_items[i]["follower_count"] = 100
            my_test_search_username_items[i]["following_count"] = 15
            my_test_search_username_items[i]["media_count"] = (
                self.bot.min_media_count_to_follow + 1
            )
            my_test_search_username_items[i]["has_anonymous_profile_picture"] = False

            my_test_username_info_items[i]["username"] = "{}_{}".format(
                TEST_FOLLOWING_ITEM["username"], i
            )
            my_test_username_info_items[i]["pk"] = TEST_FOLLOWING_ITEM["pk"] + i
            my_test_username_info_items[i]["is_verified"] = False
            my_test_username_info_items[i]["is_business"] = False
            my_test_username_info_items[i]["is_private"] = False
            my_test_username_info_items[i]["follower_count"] = 100
            my_test_username_info_items[i]["following_count"] = 15
            my_test_username_info_items[i]["media_count"] = (
                self.bot.min_media_count_to_follow + 1
            )
            my_test_username_info_items[i]["has_anonymous_profile_picture"] = False

            response_data = {"status": "ok", "user": my_test_search_username_items[i]}
            responses.add(
                responses.GET,
                "{api_url}users/{username}/usernameinfo/".format(
                    api_url=API_URL,
                    username=my_test_search_username_items[i]["username"],
                ),
                status=200,
                json=response_data,
            )

            response_data = {"status": "ok", "user": my_test_username_info_items[i]}
            responses.add(
                responses.GET,
                "{api_url}users/{user_id}/info/".format(
                    api_url=API_URL, user_id=my_test_username_info_items[i]["pk"]
                ),
                status=200,
                json=response_data,
            )

            response_data = {"status": "ok"}
            responses.add(
                responses.POST,
                "{api_url}friendships/create/{user_id}/".format(
                    api_url=API_URL, user_id=my_test_username_info_items[i]["pk"]
                ),
                json=response_data,
                status=200,
            )

        self.bot.follow_following(username)

        test_follows = self.bot.total["follows"] == follows_at_start + results_3
        test_following = sorted(self.bot.following) == [
            str(my_test_username_info_items[i]["pk"]) for i in range(results_3)
        ]
        test_followed = sorted(self.bot.followed_file.list) == [
            str(my_test_username_info_items[i]["pk"]) for i in range(results_3)
        ]
        assert test_follows and test_following and test_followed

    @responses.activate
    @pytest.mark.parametrize(
        "blocked_actions_protection,blocked_actions_sleep,result",
        [
            (True, True, False),
            (True, False, True),
            (False, True, False),
            (False, False, False),
        ],
    )
    @patch("time.sleep", return_value=None)
    def test_sleep_feedback_successful(
        self,
        patched_time_sleep,
        blocked_actions_protection,
        blocked_actions_sleep,
        result,
    ):
        self.bot.blocked_actions_protection = blocked_actions_protection
        # self.bot.blocked_actions["likes"] = False
        self.bot.blocked_actions_sleep = blocked_actions_sleep
        user_id = 1234567890
        response_data = {
            u"status": u"fail",
            u"feedback_title": u"feedback_required",
            u"feedback_message": u"This action was blocked. Please"
            u" try again later. We restrict certain content and "
            u"actions to protect our community. Tell us if you think"
            u" we made a mistake.",
            u"spam": True,
            u"feedback_action": u"report_problem",
            u"feedback_appeal_label": u"Report problem",
            u"feedback_ignore_label": u"OK",
            u"message": u"feedback_required",
            u"feedback_url": u"repute/report_problem/instagram_like_add/",
        }
        responses.add(
            responses.POST,
            "{api_url}friendships/create/{user_id}/".format(
                api_url=API_URL, user_id=user_id
            ),
            json=response_data,
            status=400,
        )
        responses.add(
            responses.POST,
            "{api_url}friendships/create/{user_id}/".format(
                api_url=API_URL, user_id=user_id
            ),
            status=200,
            json={"status": "ok"},
        )
        self.bot.follow(user_id, check_user=False)
        self.bot.follow(user_id, check_user=False)
        assert self.bot.blocked_actions["follows"] == result

    @responses.activate
    @pytest.mark.parametrize(
        "blocked_actions_protection,blocked_actions_sleep,result",
        [
            (True, True, True),
            (True, False, True),
            (False, True, False),
            (False, False, False),
        ],
    )
    @patch("time.sleep", return_value=None)
    def test_sleep_feedback_unsuccessful(
        self,
        patched_time_sleep,
        blocked_actions_protection,
        blocked_actions_sleep,
        result,
    ):
        self.bot.blocked_actions_protection = blocked_actions_protection
        # self.bot.blocked_actions["likes"] = False
        self.bot.blocked_actions_sleep = blocked_actions_sleep
        user_id = 1234567890
        response_data = {
            u"status": u"fail",
            u"feedback_title": u"feedback_required",
            u"feedback_message": u"This action was blocked. Please"
            u" try again later. We restrict certain content and "
            u"actions to protect our community. Tell us if you think"
            u" we made a mistake.",
            u"spam": True,
            u"feedback_action": u"report_problem",
            u"feedback_appeal_label": u"Report problem",
            u"feedback_ignore_label": u"OK",
            u"message": u"feedback_required",
            u"feedback_url": u"repute/report_problem/instagram_like_add/",
        }
        responses.add(
            responses.POST,
            "{api_url}friendships/create/{user_id}/".format(
                api_url=API_URL, user_id=user_id
            ),
            json=response_data,
            status=400,
        )
        responses.add(
            responses.POST,
            "{api_url}friendships/create/{user_id}/".format(
                api_url=API_URL, user_id=user_id
            ),
            json=response_data,
            status=400,
        )
        self.bot.follow(user_id, check_user=False)
        self.bot.follow(user_id, check_user=False)
        assert self.bot.blocked_actions["follows"] == result



# flake8: noqa
"""
Configuration file of the project
"""
# Config variables taken from
# https://github.com/ping/instagram_private_api/blob/master/
#   instagram_private_api/constants.py
API_DOMAIN = "i.instagram.com"
API_URL = "https://{domain}/api/v1/".format(domain=API_DOMAIN)
USER_AGENT_BASE = (
    "Instagram {instagram_version} "
    "Android ({android_version}/{android_release}; "
    "{dpi}; {resolution}; {manufacturer}; "
    "{device}; {model}; {cpu}; en_US)"
)
SIG_KEY_VERSION = "4"
# version 105.0.0.18.119 (166149665)
# August 6, 2019 from https://www.apkmirror.com/apk/instagram/
#   instagram-instagram/instagram-instagram-105-0-0-18-119-
#   release/instagram-105-0-0-18-119-4-android-apk-download/
# Check website regularly to update IG_SIG_KEY and
# X-IG-Capabilities in REQUEST_HEADERS
IG_SIG_KEY = "5f3e50f435583c9ae626302a71f7340044087a7e2c60adacfc254205a993e305"

# Request variables taken from
# https://github.com/ping/instagram_private_api/blob/
#   422d61f0a8cc9de3d5a0e78bcba53751c44e5d63/
#   instagram_private_api/client.py#L375
REQUEST_HEADERS = {
    "Connection": "Keep-Alive",
    # version 105.0.0.18.119 (166149665)
    # python3 -c "import base64, struct; print(base64.b64encode(
    #   struct.pack('<i', 166149665)).decode('ascii'));"
    "X-IG-Capabilities": "IT7nCQ==",
    "X-IG-App-ID": "567067343352427",
    "X-IG-Connection-Type": "WIFI",
    "X-IG-Prefetch-Request": "foreground",
    "X-IG-VP9-Capable": "false",
    "X-FB-HTTP-Engine": "Liger",
    "Accept": "*/*",
    "Accept-Encoding": "gzip,deflate",
    "Accept-Language": "en-US",
    "Content-type": "application/x-www-form-urlencoded; charset=UTF-8",
    "Cookie2": "$Version=1",
}

# Moved to the end as it's quite long
EXPERIMENTS = (
    "ig_android_sticker_search_explorations,android_ig_camera_ar_asset_manager"
    "_improvements_universe,ig_android_stories_seen_state_serialization,ig_sto"
    "ries_photo_time_duration_universe,ig_android_bitmap_cache_executor_size,i"
    "g_android_stories_music_search_typeahead,ig_android_delayed_comments,ig_a"
    "ndroid_switch_back_option,ig_android_video_profiler_loom_traces,ig_androi"
    "d_paid_branded_content_rendering,ig_android_direct_app_reel_grid_search,i"
    "g_android_stories_no_inflation_on_app_start,ig_android_camera_sdk_check_g"
    "l_surface_r2,ig_promote_review_screen_title_universe,ig_android_direct_ne"
    "wer_single_line_composer_universe,ig_direct_holdout_h1_2019,ig_explore_20"
    "19_h1_destination_cover,ig_android_direct_stories_in_direct_inbox,ig_fb_g"
    "raph_differentiation_no_fb_data,ig_android_recyclerview_binder_group_enab"
    "led_universe,ig_android_direct_share_sheet_custom_fast_scroller,ig_androi"
    "d_video_exoplayer_2,ig_android_shopping_channel_in_explore,ig_android_sto"
    "ries_music_filters,ig_android_2018_h1_hashtag_report_universe,ig_android_"
    "live_replay_highlights_universe,ig_android_hashtag_page_reduced_related_i"
    "tems,ig_android_live_titles_broadcaster_side_create_title_universe,ig_and"
    "roid_fbns_preload_direct_universe,ig_android_prefetch_carousels_on_swipe_"
    "universe,ig_camera_network_activity_logger,ig_camera_remove_display_rotat"
    "ion_cb_universe,ig_android_interactions_migrate_inline_composer_to_viewpo"
    "int_universe,ig_android_realtime_always_start_connection_on_condition_uni"
    "verse,ig_android_ad_leadgen_single_screen_universe,ig_android_enable_zero"
    "_rating,ig_android_import_page_post_after_biz_conversion,ig_camera_ar_eff"
    "ect_attribution_position,ig_android_vc_call_ended_cleanup_universe,ig_sto"
    "ries_engagement_holdout_2019_h1_universe,ig_android_story_import_intent,i"
    "g_direct_report_conversation_universe,ig_biz_graph_connection_universe,ig"
    "_android_codec_high_profile,ig_android_nametag,ig_android_sso_family_key_"
    "universe,ig_android_parse_direct_messages_bytes_universe,ig_hashtag_creat"
    "ion_universe,ig_android_gallery_order_by_date_taken,ig_android_igtv_resha"
    "re,ig_end_of_feed_universe,ig_android_share_others_post_reorder,ig_androi"
    "d_additional_contact_in_nux,ig_android_live_use_all_preview_sizes,ig_andr"
    "oid_clarify_invite_options,ig_android_live_align_by_2_universe,ig_android"
    "_separate_network_executor,ig_android_realtime_manager_optimization,ig_an"
    "droid_auto_advance_su_unit_when_scrolled_off_screen,ig_android_network_ca"
    "ncellation,ig_android_media_as_sticker,ig_android_stories_video_prefetch_"
    "kb,ig_android_maintabfragment,ig_inventory_connections,ig_stories_injecti"
    "on_tool_enabled_universe,ig_android_stories_disable_highlights_media_prel"
    "oading,ig_android_live_start_broadcast_optimized_universe,ig_android_stor"
    "ies_question_response_mutation_universe,ig_android_onetap_upsell_change_p"
    "wd,ig_nametag_data_collection,ig_android_disable_scroll_listeners,ig_andr"
    "oid_persistent_nux,ig_android_igtv_audio_always_on,ig_android_enable_lige"
    "r_preconnect_universe,ig_android_persistent_duplicate_notif_checker_user_"
    "based,ig_android_rate_limit_mediafeedviewablehelper,ig_android_search_rem"
    "ove_null_state_sections,ig_android_stories_viewer_drawable_cache_universe"
    ",ig_direct_android_reply_modal_universe,ig_android_biz_qp_suggest_page,ig"
    "_shopping_indicator_content_variations_android,ig_android_stories_reel_me"
    "dia_item_automatic_retry,ig_fb_notification_universe,ig_android_live_disa"
    "ble_speed_test_ui_timeout_universe,ig_android_direct_thread_scroll_perf_o"
    "ncreate_universe,ig_android_low_data_mode_backup_2,ig_android_invite_xout"
    "_universe,ig_android_low_data_mode_backup_3,ig_android_low_data_mode_back"
    "up_4,ig_android_low_data_mode_backup_5,ig_android_video_abr_universe,ig_a"
    "ndroid_low_data_mode_backup_1,ig_android_signup_refactor_santity,ig_chall"
    "enge_general_v2,ig_android_place_signature_universe,ig_android_hide_butto"
    "n_for_invite_facebook_friends,ig_android_business_promote_tooltip,ig_andr"
    "oid_follow_requests_ui_improvements,ig_android_shopping_post_tagging_nux_"
    "universe,ig_android_stories_sensitivity_screen,ig_android_camera_arengine"
    "_shader_caching_universe,ig_android_insta_video_broadcaster_infra_perf,ig"
    "_android_direct_view_more_qe,ig_android_direct_visual_message_prefetch_co"
    "unt_universe,ig_camera_android_ar_effect_stories_deeplink,ig_android_clie"
    "nt_side_delivery_universe,ig_android_stories_send_client_reels_on_tray_fe"
    "tch_universe,ig_android_direct_inbox_background_view_models,ig_android_st"
    "artup_thread_priority,ig_android_stories_viewer_responsiveness_universe,i"
    "g_android_live_use_rtc_upload_universe,ig_android_live_ama_viewer_univers"
    "e,ig_android_business_id_conversion_universe,ig_smb_ads_holdout_2018_h2_u"
    "niverse,ig_android_modal_activity_no_animation_fix_universe,ig_android_ca"
    "mera_post_smile_low_end_universe,ig_android_live_realtime_comments_univer"
    "se,ig_android_vc_in_app_notification_universe,ig_eof_caboose_universe,ig_"
    "android_new_one_tap_nux_universe,ig_android_igds_edit_profile_fields,ig_a"
    "ndroid_downgrade_viewport_exit_behavior,ig_android_mi_batch_upload_univer"
    "se,ig_camera_android_segmentation_async_universe,ig_android_use_recyclerv"
    "iew_for_direct_search_universe,ig_android_live_comment_fetch_frequency_un"
    "iverse,ig_android_create_page_on_top_universe,ig_android_direct_log_badge"
    "_count_inconsistent,ig_android_stories_text_format_emphasis,ig_android_qu"
    "estion_sticker_replied_state,ig_android_ad_connection_manager_universe,ig"
    "_android_image_upload_skip_queue_only_on_wifi,ig_android_ad_watchbrowse_c"
    "arousel_universe,ig_android_interactions_show_verified_badge_for_preview_"
    "comments_universe,ig_stories_question_sticker_music_format_prompt,ig_andr"
    "oid_activity_feed_row_click,ig_android_hide_crashing_newsfeed_story_t3813"
    "1972,ig_android_video_upload_quality_qe1,ig_android_save_collaborative_co"
    "llections,ig_android_location_attribution_text,ig_camera_android_profile_"
    "ar_notification_universe,coupon_price_test_boost_instagram_media_acquisit"
    "ion_universe,ig_android_video_outputsurface_handlerthread_universe,ig_and"
    "roid_country_code_fix_universe,ig_perf_android_holdout_2018_h1,ig_android"
    "_stories_music_overlay,ig_android_enable_lean_crash_reporting_universe,ig"
    "_android_resumable_downloads_logging_universe,ig_android_stories_default_"
    "rear_camera_universe,ig_android_low_latency_consumption_universe,ig_andro"
    "id_offline_mode_holdout,ig_android_foreground_location_collection,ig_andr"
    "oid_stories_close_friends_disable_first_time_badge,ig_android_react_nativ"
    "e_universe_kill_switch,ig_android_video_ta_universe,ig_android_media_rows"
    "_async_inflate,ig_android_stories_gallery_video_segmentation,ig_android_s"
    "tories_in_feed_preview_notify_fix_universe,ig_android_video_rebind_force_"
    "keep_playing_fix,ig_android_direct_business_holdout,ig_android_xposting_u"
    "psell_directly_after_sharing_to_story,ig_android_gallery_high_quality_pho"
    "to_thumbnails,ig_android_interactions_new_comment_like_pos_universe,ig_fe"
    "ed_core_experience_universe,ig_android_friends_sticker,ig_android_busines"
    "s_ix_universe,ig_android_suggested_highlights,ig_android_stories_posting_"
    "offline_ui,ig_android_stories_close_friends_rings_remove_green_universe,i"
    "g_android_canvas_tilt_to_pan_universe,ig_android_vc_background_call_toast"
    "_universe,ig_android_concurrent_cold_start_universe,ig_promote_default_de"
    "stination_universe,mi_viewpoint_viewability_universe,ig_android_location_"
    "page_info_page_upsell,igds_android_listrow_migration_universe,ig_direct_r"
    "eshare_sharesheet_ranking,ig_android_fb_sync_options_universe,ig_android_"
    "drawable_usage_logging_universe,ig_android_recommend_accounts_destination"
    "_routing_fix,ig_android_fix_prepare_direct_push,ig_direct_android_larger_"
    "media_reshare_style,ig_android_video_feed_universe,ig_android_building_ay"
    "mf_universe,ig_android_internal_sticker_universe,ig_traffic_routing_unive"
    "rse,ig_android_search_normalization,ig_android_ad_watchmore_entry_point_u"
    "niverse,ig_camera_android_segmentation_enabled_universe,ig_android_igtv_a"
    "lways_show_browse_ui,ig_android_page_claim_deeplink_qe,ig_explore_2018_h2"
    "_account_rec_deduplication_android,ig_android_story_accidentally_click_in"
    "vestigation_universe,ig_android_shopping_pdp_hero_carousel,ig_android_cle"
    "ar_inflight_image_request,ig_android_show_su_in_other_users_follow_list,i"
    "g_android_stories_infeed_lower_threshold_launch,ig_android_main_feed_vide"
    "o_countdown_timer,instagram_interests_holdout,ig_android_continuous_video"
    "_capture,ig_android_category_search_edit_profile,ig_android_contact_invit"
    "es_nux_universe,ig_android_settings_search_v2_universe,ig_android_video_u"
    "pload_iframe_interval,ig_business_new_value_prop_universe,ig_android_powe"
    "r_metrics,ig_android_stories_collapse_seen_segments,ig_android_live_follo"
    "w_from_comments_universe,ig_android_hashtag_discover_tab,ig_android_live_"
    "skip_live_encoder_pts_correction,ig_android_reel_zoom_universe,enable_cre"
    "ator_account_conversion_v0_universe,ig_android_test_not_signing_address_b"
    "ook_unlink_endpoint,ig_android_direct_tabbed_media_picker,ig_android_dire"
    "ct_mutation_manager_job_scheduler,ig_ei_option_setting_universe,ig_androi"
    "d_hashtag_related_items_over_logging,ig_android_livewith_liveswap_optimiz"
    "ation_universe,ig_android_direct_new_intro_card,ig_camera_android_support"
    "ed_capabilities_api_universe,ig_android_video_webrtc_textureview,ig_andro"
    "id_share_claim_page_universe,ig_direct_android_mentions_sender,ig_android"
    "_whats_app_contact_invite_universe,ig_android_video_scrubber_thumbnail_un"
    "iverse,ig_camera_ar_image_transform_library,ig_android_insights_creation_"
    "growth_universe,ig_android_igtv_refresh_tv_guide_interval,ig_android_stor"
    "ies_gif_sticker,ig_android_stories_music_broadcast_receiver,ig_android_fb"
    "_profile_integration_fbnc_universe,ig_android_low_data_mode,ig_fb_graph_d"
    "ifferentiation_control,ig_android_show_create_content_pages_universe,ig_a"
    "ndroid_igsystrace_universe,ig_android_new_contact_invites_entry_points_un"
    "iverse,ig_android_ccu_jobscheduler_inner,ig_android_netego_scroll_perf,ig"
    "_android_fb_connect_follow_invite_flow,ig_android_invite_list_button_rede"
    "sign_universe,ig_android_react_native_email_sms_settings_universe,ig_andr"
    "oid_igtv_aspect_ratio_limits,ig_hero_player,ig_android_save_auto_sharing_"
    "to_fb_option_on_server,ig_android_live_presence_universe,ig_android_white"
    "hat_options_universe,android_cameracore_preview_frame_listener2_ig_univer"
    "se,ig_android_memory_manager,ig_account_recs_in_chaining,ig_explore_2018_"
    "finite_chain_android_universe,ig_android_tagging_video_preview,ig_android"
    "_feed_survey_viewpoint,ig_android_hashtag_search_suggestions,ig_android_p"
    "rofile_neue_infra_rollout_universe,ig_android_instacrash_detection,ig_and"
    "roid_interactions_add_search_bar_to_likes_list_universe,ig_android_vc_cap"
    "ture_universe,ig_nametag_local_ocr_universe,ig_branded_content_share_to_f"
    "acebook,ig_android_direct_segmented_video,ig_android_search_page_v2,ig_an"
    "droid_stories_recently_captured_universe,ig_business_integrity_ipc_univer"
    "se,ig_android_share_product_universe,ig_fb_graph_differentiation_top_k_fb"
    "_coefficients,ig_shopping_viewer_share_action,ig_android_direct_share_sto"
    "ry_to_facebook,ig_android_business_attribute_sync,ig_android_video_time_t"
    "o_live_cache_eviction,ig_android_location_feed_related_business,ig_androi"
    "d_view_and_likes_cta_universe,ig_live_holdout_h2_2018,ig_android_profile_"
    "memories_universe,ig_promote_budget_warning_view_universe,ig_android_redi"
    "rect_to_web_on_oembed_fail_universe,ig_android_optic_new_focus_controller"
    ",ig_android_shortcuts,ig_android_search_hashtag_badges,ig_android_navigat"
    "ion_latency_logger,ig_android_direct_composer_avoid_hiding_thread_camera,"
    "ig_android_direct_remix_visual_messages,ig_android_custom_story_import_in"
    "tent,ig_android_biz_new_choose_category,ig_android_view_info_universe,ig_"
    "android_camera_upsell_dialog,ig_android_business_ix_self_serve,ig_android"
    "_dead_code_detection,ig_android_ad_watchbrowse_universe,ig_android_pbia_p"
    "roxy_profile_universe,ig_android_qp_kill_switch,ig_android_gap_rule_enfor"
    "cer_universe,ig_android_direct_delete_or_block_from_message_requests,ig_a"
    "ndroid_direct_left_aligned_navigation_bar,ig_android_feed_load_more_viewp"
    "oint_universe,ig_android_stories_reshare_reply_msg,ig_android_one_tap_sha"
    "resheet_fb_extensions,ig_android_stories_feeback_message_composer_entry_p"
    "oint,ig_direct_holdout_h2_2018,ig_camera_android_facetracker_v12_universe"
    ",ig_android_camera_ar_effects_low_storage_universe,ig_camera_android_blac"
    "k_feed_sticker_fix_universe,ig_android_direct_media_forwarding,ig_android"
    "_camera_attribution_in_direct,ig_android_audience_control,ig_android_stor"
    "ies_cross_sharing_to_fb_holdout_universe,ig_android_enable_main_feed_reel"
    "_tray_preloading,ig_android_profile_neue_universe,ig_company_profile_hold"
    "out,ig_camera_android_areffect_photo_capture_universe,ig_rti_inapp_notifi"
    "cations_universe,ig_android_vc_join_timeout_universe,ig_android_feed_core"
    "_ads_2019_h1_holdout_universe,ig_android_interactions_composer_mention_se"
    "arch_universe,ig_android_igtv_save,ig_android_follower_following_whatsapp"
    "_invite_universe,ig_android_claim_location_page,ig_android_story_ads_2019"
    "_h1_holdout_universe,ig_android_3pspp,ig_android_cache_timespan_objects,i"
    "g_timestamp_public_test,ig_android_histogram_reporter,ig_android_feed_aut"
    "o_share_to_facebook_dialog,ig_android_arengine_separate_prepare,ig_androi"
    "d_skip_button_content_on_connect_fb_universe,ig_android_igtv_profile_tab,"
    "ig_android_show_fb_name_universe,ig_android_interactions_inline_composer_"
    "extensions_universe,ig_camera_async_space_validation_for_ar,ig_android_pi"
    "geon_sampling,ig_story_camera_reverse_video_experiment,ig_android_live_us"
    "e_timestamp_normalizer,ig_android_profile_lazy_load_carousel_media,ig_and"
    "roid_stories_question_sticker_music_format,ig_business_profile_18h1_holdo"
    "ut_universe,ig_pacing_overriding_universe,ig_android_direct_allow_multili"
    "ne_composition,ig_android_interactions_emoji_extension_followup_universe,"
    "ig_android_story_ads_direct_cta_universe,ig_android_q3lc_transparency_con"
    "trol_settings,ig_stories_selfie_sticker,ig_android_sso_use_trustedapp_uni"
    "verse,ig_android_ad_increase_story_adpreload_priority_universe,ig_android"
    "_interests_netego_dismiss,ig_direct_giphy_gifs_rating,ig_android_shopping"
    "_catalogsearch,ig_android_stories_music_awareness_universe,ig_android_qcc"
    "_perf,ig_android_stories_reels_tray_media_count_check,ig_android_new_fb_p"
    "age_selection,ig_android_facebook_crosspost,ig_android_internal_collab_sa"
    "ve,ig_video_holdout_h2_2017,ig_android_story_sharing_universe,ig_promote_"
    "post_insights_entry_universe,ig_android_direct_thread_store_rewrite,ig_an"
    "droid_qp_clash_management_enabled_v4_universe,ig_branded_content_paid_bra"
    "nded_content,ig_android_large_heap_override,ig_android_live_subscribe_use"
    "r_level_universe,ig_android_igtv_creation_flow,ig_android_video_call_fini"
    "sh_universe,ig_android_direct_mqtt_send,ig_android_do_not_fetch_follow_re"
    "quests_on_success,ig_android_remove_push_notifications,ig_android_vc_dire"
    "ctapp_integration_universe,ig_android_explore_discover_people_entry_point"
    "_universe,ig_android_sonar_prober_universe,ig_android_live_bg_download_fa"
    "ce_filter_assets_universe,ig_android_gif_framerate_throttling,ig_android_"
    "live_webrtc_livewith_params,ig_android_vc_always_start_connection_on_cond"
    "ition_universe,ig_camera_worldtracking_set_scale_by_arclass,ig_android_di"
    "rect_inbox_typing_indicator,ig_android_stories_music_lyrics_scrubber,ig_f"
    "eed_experience,ig_android_direct_new_thread_local_search_fix_universe,ig_"
    "android_appstate_logger,ig_promote_insights_video_views_universe,ig_andro"
    "id_dismiss_recent_searches,ig_android_downloadable_igrtc_module,ig_androi"
    "d_fb_link_ui_polish_universe,ig_stories_music_sticker,ig_android_device_c"
    "apability_framework,ig_scroll_by_two_cards_for_suggested_invite_universe,"
    "ig_android_stories_helium_balloon_badging_universe,ig_android_business_re"
    "move_unowned_fb_pages,ig_android_stories_combined_asset_search,ig_stories"
    "_allow_camera_actions_while_recording,ig_android_analytics_mark_events_as"
    "_offscreen,ig_android_optic_feature_testing,ig_android_camera_universe,ig"
    "_android_optic_photo_cropping_fixes,ig_camera_regiontracking_use_similari"
    "ty_tracker_for_scaling,ig_android_refreshable_list_view_check_spring,feli"
    "x_android_video_quality,ig_android_biz_endpoint_switch,ig_android_direct_"
    "continuous_capture,ig_android_comments_direct_reply_to_author,ig_android_"
    "vc_webrtc_params,ig_android_claim_or_connect_page_on_xpost,ig_android_anr"
    ",ig_android_optic_new_architecture,ig_android_stories_viewer_as_modal_hig"
    "h_end_launch,ig_android_hashtag_follow_chaining_over_logging,ig_new_eof_d"
    "emarcator_universe,ig_android_push_notifications_settings_redesign_univer"
    "se,ig_hashtag_display_universe,ig_fbns_push,coupon_price_test_ad4ad_insta"
    "gram_resurrection_universe,ig_android_live_rendering_looper_universe,ig_a"
    "ndroid_mqtt_cookie_auth_memcache_universe,ig_android_live_end_redirect_un"
    "iverse,ig_android_direct_mutation_manager_media_2,ig_android_ccu_jobsched"
    "uler_outer,ig_smb_ads_holdout_2019_h1_universe,ig_fb_graph_differentiatio"
    "n,ig_android_stories_share_extension_video_segmentation,ig_android_intera"
    "ctions_realtime_typing_indicator_and_live_comments,ig_android_stories_cre"
    "ate_flow_favorites_tooltip,ig_android_live_nerd_stats_universe,ig_android"
    "_universe_video_production,ig_android_hide_reset_with_fb_universe,ig_andr"
    "oid_reactive_feed_like_count,ig_android_stories_music_precapture,ig_andro"
    "id_vc_service_crash_fix_universe,ig_android_shopping_product_overlay,ig_a"
    "ndroid_direct_double_tap_to_like_hearts,ig_camera_android_api_rewrite_uni"
    "verse,ig_android_growth_fci_team_holdout_universe,ig_android_stories_gall"
    "ery_recyclerview_kit_universe,ig_android_story_ads_instant_sub_impression"
    "_universe,ig_business_signup_biz_id_universe,ig_android_save_all,ig_andro"
    "id_main_feed_fragment_scroll_timing_histogram_uni,ig_android_ttcp_improve"
    "ments,ig_android_camera_ar_platform_profile_universe,ig_explore_2018_topi"
    "c_channel_navigation_android_universe,ig_android_live_fault_tolerance_uni"
    "verse,ig_android_stories_viewer_tall_android_cap_media_universe,native_co"
    "ntact_invites_universe,ig_android_dash_script,ig_android_insights_media_h"
    "ashtag_insight_universe,ig_camera_fast_tti_universe,ig_android_stories_wh"
    "atsapp_share,ig_android_inappnotification_rootactivity_tweak,ig_android_r"
    "ender_thread_memory_leak_holdout,ig_android_private_highlights_universe,i"
    "g_android_rate_limit_feed_video_module,ig_android_one_tap_fbshare,ig_shar"
    "e_to_story_toggle_include_shopping_product,ig_android_direct_speed_cam_un"
    "iv,ig_payments_billing_address,ig_android_ufiv3_holdout,ig_android_new_ca"
    "mera_design_container_animations_universe,ig_android_livewith_guest_adapt"
    "ive_camera_universe,ig_android_direct_fix_playing_invalid_visual_message,"
    "ig_shopping_viewer_intent_actions,ig_promote_add_payment_navigation_unive"
    "rse,ig_android_optic_disable_post_capture_preview_restart,ig_android_main"
    "_feed_refresh_style_universe,ig_android_live_analytics,ig_android_story_a"
    "ds_performance_universe_1,ig_android_stories_viewer_modal_activity,ig_and"
    "roid_story_ads_performance_universe_3,ig_android_story_ads_performance_un"
    "iverse_4,ig_android_feed_seen_state_with_view_info,ig_android_ads_profile"
    "_cta_feed_universe,ig_android_vc_cowatch_universe,ig_android_optic_thread"
    "_priorities,ig_android_igtv_chaining,ig_android_live_qa_viewer_v1_univers"
    "e,ig_android_stories_show_story_not_available_error_msg,ig_android_inline"
    "_notifications_recommended_user,ig_shopping_post_insights,ig_android_webr"
    "tc_streamid_salt_universe,ig_android_wellbeing_timeinapp_v1_universe,ig_a"
    "ndroid_profile_cta_v3,ig_android_video_qp_logger_universe,ig_android_cach"
    "e_video_autoplay_checker,ig_android_live_suggested_live_expansion,ig_andr"
    "oid_vc_start_from_direct_inbox_universe,ig_perf_android_holdout,ig_fb_gra"
    "ph_differentiation_only_fb_candidates,ig_android_expired_build_lockout,ig"
    "_promote_lotus_universe,ig_android_video_streaming_upload_universe,ig_and"
    "roid_optic_fast_preview_restart_listener,ig_interactions_h1_2019_team_hol"
    "dout_universe,ig_android_ad_async_ads_universe,ig_camera_android_effect_i"
    "nfo_bottom_sheet_universe,ig_android_stories_feedback_badging_universe,ig"
    "_android_sorting_on_self_following_universe,ig_android_edit_location_page"
    "_info,ig_promote_are_you_sure_universe,ig_android_interactions_feed_label"
    "_below_comments_refactor_universe,ig_android_camera_platform_effect_share"
    "_universe,ig_stories_engagement_swipe_animation_simple_universe,ig_login_"
    "activity,ig_android_direct_quick_replies,ig_android_fbns_optimization_uni"
    "verse,ig_android_stories_alignment_guides_universe,ig_android_rn_ads_mana"
    "ger_universe,ig_explore_2018_post_chaining_account_recs_dedupe_universe,i"
    "g_android_click_to_direct_story_reaction_universe,ig_internal_research_se"
    "ttings,ig_android_stories_video_seeking_audio_bug_fix,ig_android_insights"
    "_holdout,ig_android_swipe_up_area_universe,ig_android_rendering_controls,"
    "ig_android_feed_post_sticker,ig_android_inline_editing_local_prefill,ig_a"
    "ndroid_hybrid_bitmap_v3_prenougat,ig_android_cronet_stack,ig_android_enab"
    "le_igrtc_module,ig_android_scroll_audio_priority,ig_android_shopping_prod"
    "uct_appeals_universe,ig_android_fb_follow_server_linkage_universe,ig_andr"
    "oid_fblocation_universe,ig_android_direct_updated_story_reference_ui,ig_c"
    "amera_holdout_h1_2018_product,live_with_request_to_join_button_universe,i"
    "g_android_music_continuous_capture,ig_android_churned_find_friends_redire"
    "ct_to_discover_people,ig_android_main_feed_new_posts_indicator_universe,i"
    "g_vp9_hd_blacklist,ig_ios_queue_time_qpl_universe,ig_android_split_contac"
    "ts_list,ig_android_connect_owned_page_universe,ig_android_felix_prefetch_"
    "thumbnail_sprite_sheet,ig_android_multi_dex_class_loader_v2,ig_android_wa"
    "tch_and_more_redesign,igtv_feed_previews,ig_android_qp_batch_fetch_cachin"
    "g_enabled_v1_universe,ig_android_profile_edit_phone_universe,ig_android_v"
    "c_renderer_type_universe,ig_android_local_2018_h2_holdout,ig_android_purx"
    "_native_checkout_universe,ig_android_vc_disable_lock_screen_content_acces"
    "s_universe,ig_android_business_transaction_in_stories_creator,android_cam"
    "eracore_ard_ig_integration,ig_video_experimental_encoding_consumption_uni"
    "verse,ig_android_iab_autofill,ig_android_location_page_intent_survey,ig_c"
    "amera_android_segmentation_qe2_universe,ig_android_image_mem_cache_strong"
    "_ref_universe,ig_android_business_promote_refresh_fb_access_token_univers"
    "e,ig_android_stories_samsung_sharing_integration,ig_android_hashtag_heade"
    "r_display,ig_discovery_holdout_2019_h1_universe,ig_android_user_url_deepl"
    "ink_fbpage_endpoint,ig_android_direct_mutation_manager_handler_thread_uni"
    "verse,ig_branded_content_show_settings_universe,ig_android_ad_holdout_wat"
    "chandmore_universe,ig_android_direct_thread_green_dot_presence_universe,i"
    "g_android_camera_new_post_smile_universe,ig_android_shopping_signup_redes"
    "ign_universe,ig_android_vc_missed_call_notification_action_reply,allow_pu"
    "blish_page_universe,ig_android_experimental_onetap_dialogs_universe,ig_pr"
    "omote_ppe_v2_universe,android_cameracore_ig_gl_oom_fixes_universe,ig_andr"
    "oid_multi_capture_camera,ig_android_fb_family_navigation_badging_user,ig_"
    "android_follow_requests_copy_improvements,ig_media_geo_gating,ig_android_"
    "comments_notifications_universe,ig_android_render_output_surface_timeout_"
    "universe,ig_android_drop_frame_check_paused,ig_direct_raven_sharesheet_ra"
    "nking,ig_android_realtime_mqtt_logging,ig_family_bridges_holdout_universe"
    ",ig_android_rainbow_hashtags,ig_android_ad_watchinstall_universe,ig_andro"
    "id_ad_account_top_followers_universe,ig_android_betamap_universe,ig_andro"
    "id_video_ssim_report_universe,ig_android_cache_network_util,ig_android_le"
    "ak_detector_upload_universe,ig_android_carousel_prefetch_bumping,ig_fbns_"
    "preload_default,ig_android_inline_appeal_show_new_content,ig_fbns_kill_sw"
    "itch,ig_hashtag_following_holdout_universe,ig_android_show_weekly_ci_upse"
    "ll_limit,ig_android_direct_reel_options_entry_point_2_universe,enable_cre"
    "ator_account_conversion_v0_animation,ig_android_http_service_same_thread,"
    "ig_camera_holdout_h1_2018_performance,ig_android_direct_mutation_manager_"
    "cancel_fix_universe,ig_music_dash,ig_android_fb_url_universe,ig_android_r"
    "eel_raven_video_segmented_upload_universe,ig_android_promote_native_migra"
    "tion_universe,ig_camera_android_badge_face_effects_universe,ig_android_hy"
    "brid_bitmap_v3_nougat,ig_android_multi_author_story_reshare_universe,ig_a"
    "ndroid_vc_camera_zoom_universe,ig_android_enable_request_compression_ccu,"
    "ig_android_video_controls_universe,ig_android_logging_metric_universe_v2,"
    "ig_android_xposting_newly_fbc_people,ig_android_visualcomposer_inapp_noti"
    "fication_universe,ig_android_contact_point_upload_rate_limit_killswitch,i"
    "g_android_webrtc_encoder_factory_universe,ig_android_search_impression_lo"
    "gging,ig_android_handle_username_in_media_urls_universe,ig_android_sso_ko"
    "totoro_app_universe,ig_android_mi_holdout_h1_2019,ig_android_igtv_autopla"
    "y_on_prepare,ig_file_based_session_handler_2_universe,ig_branded_content_"
    "tagging_upsell,ig_shopping_insights_parity_universe_android,ig_android_li"
    "ve_ama_universe,ig_android_external_gallery_import_affordance,ig_android_"
    "updatelistview_on_loadmore,ig_android_optic_new_zoom_controller,ig_androi"
    "d_hide_type_mode_camera_button,ig_android_photos_qpl,ig_android_reel_impr"
    "esssion_cache_key_qe_universe,ig_android_show_profile_picture_upsell_in_r"
    "eel_universe,ig_android_live_viewer_tap_to_hide_chrome_universe,ig_discov"
    "ery_holdout_universe,ig_android_direct_import_google_photos2,ig_android_s"
    "tories_tray_in_viewer,ig_android_request_verification_badge,ig_android_di"
    "rect_unlimited_raven_replays_inthreadsession_fix,ig_android_netgo_cta,ig_"
    "android_viewpoint_netego_universe,ig_android_stories_separate_overlay_cre"
    "ation,ig_android_iris_improvements,ig_android_biz_conversion_naming_test,"
    "ig_android_fci_empty_feed_friend_search,ig_android_hashtag_page_support_p"
    "laces_tab,ig_camera_android_ar_platform_universe,ig_android_stories_viewe"
    "r_prefetch_improvements,ig_android_optic_camera_warmup,ig_android_place_s"
    "earch_profile_image,ig_android_interactions_in_feed_comment_view_universe"
    ",ig_android_fb_sharing_shortcut,ig_android_oreo_hardware_bitmap,ig_androi"
    "d_analytics_diagnostics_universe,ig_android_insights_creative_tutorials_u"
    "niverse,ig_android_vc_universe,ig_android_profile_unified_follow_view,ig_"
    "android_collect_os_usage_events_universe,ig_android_shopping_nux_timing_u"
    "niverse,ig_android_fbpage_on_profile_side_tray,ig_android_native_logcat_i"
    "nterceptor,ig_android_direct_thread_content_picker,ig_android_notif_impro"
    "vement_universe,ig_face_effect_ranking,ig_android_shopping_more_from_busi"
    "ness,ig_feed_content_universe,ig_android_hacked_account_reporting,ig_andr"
    "oid_disk_usage_logging_universe,ig_android_ad_redesign_iab_universe,ig_an"
    "droid_banyan_migration,ig_android_profile_event_leak_holdout,ig_android_s"
    "tories_loading_automatic_retry,ig_android_gqls_typing_indicator,ag_family"
    "_bridges_2018_h2_holdout,ig_promote_net_promoter_score_universe,ig_androi"
    "d_direct_last_seen_message_indicator,ig_android_biz_conversion_suggest_bi"
    "z_nux,ig_android_log_mediacodec_info,ig_android_vc_participant_state_call"
    "ee_universe,ig_camera_android_boomerang_attribution_universe,ig_android_s"
    "tories_weblink_creation,ig_android_horizontal_swipe_lfd_logging,ig_profil"
    "e_company_holdout_h2_2018,ig_android_ads_manager_pause_resume_ads_univers"
    "e,ig_promote_fix_expired_fb_accesstoken_android_universe,ig_android_stori"
    "es_media_seen_batching_universe,ig_android_interactions_nav_to_permalink_"
    "followup_universe,ig_android_live_titles_viewer_side_view_title_universe,"
    "ig_android_direct_mark_as_read_notif_action,ig_android_edit_highlight_red"
    "esign,ig_android_direct_mutation_manager_backoff_universe,ig_android_inte"
    "ractions_comment_like_for_all_feed_universe,ig_android_mi_skip_analytic_e"
    "vent_pool_universe,ig_android_fbc_upsell_on_dp_first_load,ig_android_audi"
    "o_ingestion_params,ig_android_video_call_participant_state_caller_univers"
    "e,ig_fbns_shared,ig_feed_engagement_holdout_2018_h1,ig_camera_android_bg_"
    "processor,ig_android_optic_new_features_implementation,ig_android_stories"
    "_reel_interactive_tap_target_size,ig_android_video_live_trace_universe,ig"
    "_android_igtv_browse_with_pip_v2,ig_android_interactive_listview_during_r"
    "efresh,ig_android_igtv_feed_banner_universe,ig_android_unfollow_from_main"
    "_feed_v2,ig_android_self_story_setting_option_in_menu,ig_android_ad_watch"
    "lead_universe,ufi_share,ig_android_live_special_codec_size_list,ig_androi"
    "d_live_qa_broadcaster_v1_universe,ig_android_hide_stories_viewer_list_uni"
    "verse,ig_android_direct_albums,ig_android_business_transaction_in_stories"
    "_consumer,ig_android_scroll_stories_tray_to_front_when_stories_ready,ig_a"
    "ndroid_direct_thread_composer,instagram_android_stories_sticker_tray_rede"
    "sign,ig_camera_android_superzoom_icon_position_universe,ig_android_busine"
    "ss_cross_post_with_biz_id_infra,ig_android_photo_invites,ig_android_reel_"
    "tray_item_impression_logging_viewpoint,ig_account_identity_2018_h2_lockdo"
    "wn_phone_global_holdout,ig_android_high_res_gif_stickers,ig_close_friends"
    "_v4,ig_fb_cross_posting_sender_side_holdout,ig_android_ads_history_univer"
    "se,ig_android_comments_composer_newline_universe,ig_rtc_use_dtls_srtp,ig_"
    "promote_media_picker_universe,ig_android_live_start_live_button_universe,"
    "ig_android_vc_ongoing_call_notification_universe,ig_android_rate_limit_fe"
    "ed_item_viewable_helper,ig_android_bitmap_attribution_check,ig_android_ig"
    "_to_fb_sync_universe,ig_android_reel_viewer_data_buffer_size,ig_two_fac_t"
    "otp_enable,ig_android_vc_missed_call_notification_action_call_back,ig_and"
    "roid_stories_landscape_mode,ig_android_ad_view_ads_native_universe,ig_and"
    "roid_igtv_whitelisted_for_web,ig_android_global_prefetch_scheduler,ig_and"
    "roid_live_thread_delay_for_mute_universe,ig_close_friends_v4_global,ig_an"
    "droid_share_publish_page_universe,ig_android_new_camera_design_universe,i"
    "g_direct_max_participants,ig_promote_hide_local_awareness_universe,ig_and"
    "roid_graphql_survey_new_proxy_universe,ig_android_fs_creation_flow_tweaks"
    ",ig_android_ad_watchbrowse_cta_universe,ig_android_camera_new_tray_behavi"
    "or_universe,ig_android_direct_expiring_media_loading_errors,ig_android_sh"
    "ow_fbunlink_button_based_on_server_data,ig_android_downloadable_vp8_modul"
    "e,ig_android_igtv_feed_trailer,ig_android_fb_profile_integration_universe"
    ",ig_android_profile_private_banner,ig_camera_android_focus_attribution_un"
    "iverse,ig_android_rage_shake_whitelist,ig_android_su_follow_back,ig_andro"
    "id_prefetch_notification_data,ig_android_webrtc_icerestart_on_failure_uni"
    "verse,ig_android_vpvd_impressions_universe,ig_android_payload_based_sched"
    "uling,ig_android_grid_cell_count,ig_android_new_highlight_button_text,ig_"
    "android_direct_search_bar_redesign,ig_android_hashtag_row_preparer,ig_and"
    "roid_ad_pbia_header_click_universe,ig_android_direct_visual_viewer_ppr_fi"
    "x,ig_background_prefetch,ig_camera_android_focus_in_post_universe,ig_andr"
    "oid_time_spent_dashboard,ig_android_direct_vm_activity_sheet,ig_promote_p"
    "olitical_ads_universe,ig_android_stories_auto_retry_reels_media_and_segme"
    "nts,ig_android_recommend_accounts_killswitch,ig_shopping_video_half_sheet"
    ",ig_android_ad_iab_qpl_kill_switch_universe,ig_android_interactions_direc"
    "t_share_comment_universe,ig_android_vc_sounds_universe,ig_camera_android_"
    "cache_format_picker_children,ig_android_post_live_expanded_comments_view_"
    "universe,ig_android_always_use_server_recents,ig_android_qp_slot_cooldown"
    "_enabled_universe,ig_android_asset_picker_improvements,ig_android_direct_"
    "activator_cards,ig_android_pending_media_manager_init_fix_universe,ig_and"
    "roid_facebook_global_state_sync_frequency_universe,ig_android_network_tra"
    "ce_migration,ig_android_creation_new_post_title,ig_android_reverse_audio,"
    "ig_android_camera_gallery_upload_we_universe,ig_android_direct_inbox_asyn"
    "c_diffing_universe,ig_android_live_save_to_camera_roll_limit_by_screen_si"
    "ze_universe,ig_android_profile_phone_autoconfirm_universe,ig_direct_stori"
    "es_questions,ig_android_optic_surface_texture_cleanup,ig_android_vc_use_t"
    "imestamp_normalizer,ig_android_post_recs_show_more_button_universe,ig_sho"
    "pping_checkout_mvp_experiment,ig_android_direct_pending_media,ig_android_"
    "scroll_main_feed,ig_android_intialization_chunk_410,ig_android_story_ads_"
    "default_long_video_duration,ig_android_interactions_mention_search_presen"
    "ce_dot_universe,ig_android_stories_music_sticker_position,ig_android_dire"
    "ct_character_limit,ig_stories_music_themes,ig_android_nametag_save_experi"
    "ment_universe,ig_android_media_rows_prepare_10_31,ig_android_fs_new_galle"
    "ry,ig_android_stories_hide_retry_button_during_loading_launch,ig_android_"
    "remove_follow_all_fb_list,ig_android_biz_conversion_editable_profile_revi"
    "ew_universe,ig_android_shopping_checkout_mvp,ig_android_local_info_page,i"
    "g_android_direct_log_badge_count"
)

LOGIN_EXPERIMENTS = (
    "ig_android_fci_onboarding_friend_search,ig_android_device_detection_info_"
    "upload,ig_android_autosubmit_password_recovery_universe,ig_growth_android"
    "_profile_pic_prefill_with_fb_pic_2,ig_account_identity_logged_out_signals"
    "_global_holdout_universe,ig_android_background_voice_phone_confirmation_p"
    "refilled_phone_number_only,ig_android_login_identifier_fuzzy_match,ig_and"
    "roid_one_tap_aymh_redesign_universe,ig_android_keyboard_detector_fix,ig_a"
    "ndroid_suma_landing_page,ig_android_direct_main_tab_universe,ig_android_a"
    "ymh_signal_collecting_kill_switch,ig_android_login_forgot_password_univer"
    "se,ig_android_smartlock_hints_universe,ig_android_smart_prefill_killswitc"
    "h,ig_android_account_switch_infra_universe,ig_android_multi_tap_login_new"
    ",ig_android_email_one_tap_auto_login_during_reg,ig_android_category_searc"
    "h_in_sign_up,ig_android_report_nux_completed_device,ig_android_reg_login_"
    "profile_photo_universe,ig_android_caption_typeahead_fix_on_o_universe,ig_"
    "android_nux_add_email_device,ig_android_ci_opt_in_placement,ig_android_re"
    "member_password_at_login,ig_type_ahead_recover_account,ig_android_analyti"
    "cs_accessibility_event,ig_sem_resurrection_logging,ig_android_abandoned_r"
    "eg_flow,ig_android_editable_username_in_reg,ig_android_account_recovery_a"
    "uto_login,ig_android_sim_info_upload,ig_android_skip_signup_from_one_tap_"
    "if_no_fb_sso,ig_android_hide_fb_flow_in_add_account_flow,ig_android_mobil"
    "e_http_flow_device_universe,ig_account_recovery_via_whatsapp_universe,ig_"
    "android_hide_fb_button_when_not_installed_universe,ig_prioritize_user_inp"
    "ut_on_switch_to_signup,ig_android_gmail_oauth_in_reg,ig_android_login_saf"
    "etynet,ig_android_gmail_autocomplete_account_over_one_tap,ig_android_back"
    "ground_voice_phone_confirmation,ig_android_phone_auto_login_during_reg,ig"
    "_android_hide_typeahead_for_logged_users,ig_android_hindi,ig_android_reg_"
    "modularization_universe,ig_android_bottom_sheet,ig_android_snack_bar_hidi"
    "ng,ig_android_one_tap_fallback_auto_login,ig_android_device_verification_"
    "separate_endpoint,ig_account_recovery_with_code_android_universe,ig_andro"
    "id_onboarding_skip_fb_connect,ig_android_phone_reg_redesign_universe,ig_a"
    "ndroid_universe_noticiation_channels,ig_android_media_cache_cleared_unive"
    "rse,ig_android_account_linking_universe,ig_android_hsite_prefill_new_carr"
    "ier,ig_android_retry_create_account_universe,ig_android_family_apps_user_"
    "values_provider_universe,ig_android_reg_nux_headers_cleanup_universe,ig_a"
    "ndroid_dialog_email_reg_error_universe,ig_android_ci_fb_reg,ig_android_de"
    "vice_info_foreground_reporting,ig_fb_invite_entry_points,ig_android_devic"
    "e_verification_fb_signup,ig_android_suma_biz_account,ig_android_onetaplog"
    "in_optimization,ig_video_debug_overlay,ig_android_ask_for_permissions_on_"
    "reg,ig_android_display_full_country_name_in_reg_universe,ig_android_exopl"
    "ayer_settings,ig_android_persistent_duplicate_notif_checker,ig_android_se"
    "curity_intent_switchoff,ig_android_background_voice_confirmation_block_ar"
    "gentinian_numbers,ig_android_do_not_show_back_button_in_nux_user_list,ig_"
    "android_passwordless_auth,ig_android_direct_main_tab_account_switch,ig_an"
    "droid_modularized_dynamic_nux_universe,ig_android_icon_perf2,ig_android_e"
    "mail_suggestions_universe,ig_android_fb_account_linking_sampling_freq_uni"
    "verse,ig_android_prefill_full_name_from_fb,ig_android_access_flow_prefill"
)

LAUNCHER_CONFIGS = (
    "ig_android_felix_release_players,ig_user_mismatch_soft_error,"
    "ig_android_os_version_blocking_config,"
    "ig_android_carrier_signals_killswitch,fizz_ig_android,"
    "ig_mi_block_expired_events,ig_android_killswitch_perm_direct_ssim,"
    "ig_fbns_blocked"
)

SUPPORTED_CAPABILITIES = [
    {
        "name": "SUPPORTED_SDK_VERSIONS",
        "value": "13.0,14.0,15.0,16.0,17.0,18.0,19.0,20.0,21.0,22.0,23.0,24."
        + "0,25.0,26.0,27.0,28.0,29.0,30.0,31.0,32.0,33.0,34.0,35.0,36.0,37.0,"
        + "38.0,39.0,40.0,41.0,42.0,43.0,44.0,45.0,46.0,47.0,48.0,49.0,50.0,51"
        + ".0,52.0,53.0",
    },
    {"name": "FACE_TRACKER_VERSION", "value": "12"},
    {"name": "segmentation", "value": "segmentation_enabled"},
    {"name": "WORLD_TRACKER", "value": "WORLD_TRACKER_ENABLED"},
]



import datetime

from instabot.singleton import Singleton


class BotState(object):
    __metaclass__ = Singleton

    def __init__(self):
        self.start_time = datetime.datetime.now()
        self.total = dict.fromkeys(
            [
                "likes",
                "unlikes",
                "follows",
                "unfollows",
                "comments",
                "blocks",
                "unblocks",
                "messages",
                "archived",
                "unarchived",
                "stories_viewed",
            ],
            0,
        )
        self.blocked_actions = dict.fromkeys(
            [
                "likes",
                "unlikes",
                "follows",
                "unfollows",
                "comments",
                "blocks",
                "unblocks",
                "messages",
            ],
            False,
        )
        self.sleeping_actions = dict.fromkeys(
            [
                "likes",
                "unlikes",
                "follows",
                "unfollows",
                "comments",
                "blocks",
                "unblocks",
                "messages",
            ],
            False,
        )
        self.last = dict.fromkeys(
            [
                "like",
                "unlike",
                "follow",
                "unfollow",
                "comment",
                "block",
                "unblock",
                "message",
            ],
            0,
        )

    def __repr__(self):
        return self.__dict__



import json
import os
import random
import time
import traceback

import requests
import requests.utils

from . import config, devices

# ====== SYNC METHODS ====== #


def sync_device_features(self, login=False):
    data = {
        "id": self.uuid,
        "server_config_retrieval": "1",
        "experiments": config.LOGIN_EXPERIMENTS,
    }
    if login is False:
        data["_uuid"] = self.uuid
        data["_uid"] = self.user_id
        data["_csrftoken"] = self.token
    data = json.dumps(data)
    return self.send_request(
        "qe/sync/", data, login=login, headers={"X-DEVICE-ID": self.uuid}
    )


def sync_launcher(self, login=False):
    data = {
        "id": self.uuid,
        "server_config_retrieval": "1",
        "experiments": config.LAUNCHER_CONFIGS,
    }
    if login is False:
        data["_uuid"] = self.uuid
        data["_uid"] = self.user_id
        data["_csrftoken"] = self.token
    data = json.dumps(data)
    return self.send_request("launcher/sync/", data, login=login)


def sync_user_features(self):
    data = self.default_data
    data["id"] = self.uuid
    data["experiments"] = config.EXPERIMENTS
    data = json.dumps(data)
    self.last_experiments = time.time()
    return self.send_request("qe/sync/", data, headers={"X-DEVICE-ID": self.uuid})


# ====== LOGIN/PRE FLOWS METHODS ====== #


def pre_login_flow(self):
    self.logger.info("PRE-LOGIN FLOW!... ")

    self.read_msisdn_header("default")
    self.sync_launcher(True)
    self.sync_device_features(True)
    self.log_attribution()
    self.set_contact_point_prefill("prefill")


def login_flow(self, just_logged_in=False, app_refresh_interval=1800):
    self.logger.info("LOGIN FLOW! Just logged-in: {}".format(just_logged_in))
    check_flow = []
    if just_logged_in:
        try:
            # SYNC
            check_flow.append(self.sync_launcher(False))
            check_flow.append(self.sync_user_features())
            # Update feed and timeline
            check_flow.append(self.get_timeline_feed())
            check_flow.append(self.get_reels_tray_feed(reason="cold_start"))
            check_flow.append(self.get_suggested_searches("users"))
            # getRecentSearches() ...
            check_flow.append(self.get_suggested_searches("blended"))
            # DM-Update
            check_flow.append(self.get_ranked_recipients("reshare", True))
            check_flow.append(self.get_ranked_recipients("save", True))
            check_flow.append(self.get_inbox_v2())
            check_flow.append(self.get_presence())
            check_flow.append(self.get_recent_activity())
            # Config and other stuffs
            check_flow.append(self.get_loom_fetch_config())
            check_flow.append(self.get_profile_notice())
            check_flow.append(self.batch_fetch())
            # getBlockedMedia() ...
            check_flow.append(self.explore(True))
            # getQPFetch() ...
            # getFacebookOTA() ...
        except Exception as e:
            self.logger.error(
                "Exception raised: {}\n{}".format(e, traceback.format_exc())
            )
            return False
    else:
        try:
            pull_to_refresh = random.randint(1, 100) % 2 == 0
            check_flow.append(
                self.get_timeline_feed(
                    options=["is_pull_to_refresh"] if pull_to_refresh is True else []
                )
            )  # Random pull_to_refresh :)
            check_flow.append(
                self.get_reels_tray_feed(
                    reason="pull_to_refresh"
                    if pull_to_refresh is True
                    else "cold_start"
                )
            )

            is_session_expired = (time.time() - self.last_login) > app_refresh_interval
            if is_session_expired:
                self.last_login = time.time()
                self.client_session_id = self.generate_UUID(uuid_type=True)

                # getBootstrapUsers() ...
                check_flow.append(self.get_ranked_recipients("reshare", True))
                check_flow.append(self.get_ranked_recipients("save", True))
                check_flow.append(self.get_inbox_v2())
                check_flow.append(self.get_presence())
                check_flow.append(self.get_recent_activity())
                check_flow.append(self.get_profile_notice())
                check_flow.append(self.explore(False))

            if (time.time() - self.last_experiments) > 7200:
                check_flow.append(self.sync_user_features())
                check_flow.append(self.sync_device_features())
        except Exception as e:
            self.logger.error(
                "Exception raised: {}\n{}".format(e, traceback.format_exc())
            )
            return False

    self.save_uuid_and_cookie()
    return False if False in check_flow else True


# ====== DEVICE / CLIENT_ID / PHONE_ID AND OTHER VALUES (uuids) ====== #


def set_device(self):
    self.device_settings = devices.DEVICES[self.device]
    self.user_agent = config.USER_AGENT_BASE.format(**self.device_settings)


def generate_all_uuids(self):
    self.phone_id = self.generate_UUID(uuid_type=True)
    self.uuid = self.generate_UUID(uuid_type=True)
    self.client_session_id = self.generate_UUID(uuid_type=True)
    self.advertising_id = self.generate_UUID(uuid_type=True)
    self.device_id = self.generate_device_id(
        self.get_seed(self.username, self.password)
    )


def reinstall_app_simulation(self):
    self.logger.info("Reinstall app simulation, generating new `phone_id`...")
    self.phone_id = self.generate_UUID(uuid_type=True)
    self.save_uuid_and_cookie()
    self.logger.info("New phone_id: {}".format(self.phone_id))


def change_device_simulation(self):
    self.logger.info("Change device simulation")
    self.reinstall_app_simulation()
    self.logger.info("Generating new `android_device_id`...")
    self.device_id = self.generate_device_id(
        self.get_seed(self.generate_UUID(uuid_type=True))
    )
    self.save_uuid_and_cookie()
    self.logger.info("New android_device_id: {}".format(self.device_id))


def load_uuid_and_cookie(self, load_uuid=True, load_cookie=True):
    if self.cookie_fname is None:
        fname = "{}_uuid_and_cookie.json".format(self.username)
        self.cookie_fname = os.path.join(self.base_path, fname)

    if os.path.isfile(self.cookie_fname) is False:
        return False

    with open(self.cookie_fname, "r") as f:
        data = json.load(f)
        if "cookie" in data:
            self.last_login = data["timing_value"]["last_login"]
            self.last_experiments = data["timing_value"]["last_experiments"]

            if load_cookie:
                self.logger.debug("Loading cookies")
                self.session.cookies = requests.utils.cookiejar_from_dict(
                    data["cookie"]
                )
                cookie_username = self.cookie_dict["ds_user"]
                assert cookie_username == self.username.lower()

            if load_uuid:
                self.logger.debug("Loading uuids")
                self.phone_id = data["uuids"]["phone_id"]
                self.uuid = data["uuids"]["uuid"]
                self.client_session_id = data["uuids"]["client_session_id"]
                self.advertising_id = data["uuids"]["advertising_id"]
                self.device_id = data["uuids"]["device_id"]

                self.device_settings = data["device_settings"]
                self.user_agent = data["user_agent"]

            msg = (
                "Recovery from {}: COOKIE {} - UUIDs {} - TIMING, DEVICE "
                "and ...\n- user-agent={}\n- phone_id={}\n- uuid={}\n- "
                "client_session_id={}\n- device_id={}"
            )

            self.logger.info(
                msg.format(
                    self.cookie_fname,
                    load_cookie,
                    load_uuid,
                    self.user_agent,
                    self.phone_id,
                    self.uuid,
                    self.client_session_id,
                    self.device_id,
                )
            )
        else:
            self.logger.info(
                "The cookie seems to be the with the older structure. "
                "Load and init again all uuids"
            )
            self.session.cookies = requests.utils.cookiejar_from_dict(data)
            self.last_login = time.time()
            self.last_experiments = time.time()
            cookie_username = self.cookie_dict["ds_user"]
            assert cookie_username == self.username
            self.set_device()
            self.generate_all_uuids()

    self.is_logged_in = True
    return True


def save_uuid_and_cookie(self):
    if self.cookie_fname is None:
        fname = "{}_uuid_and_cookie.json".format(self.username)
        self.cookie_fname = os.path.join(self.base_path, fname)

    data = {
        "uuids": {
            "phone_id": self.phone_id,
            "uuid": self.uuid,
            "client_session_id": self.client_session_id,
            "advertising_id": self.advertising_id,
            "device_id": self.device_id,
        },
        "cookie": requests.utils.dict_from_cookiejar(self.session.cookies),
        "timing_value": {
            "last_login": self.last_login,
            "last_experiments": self.last_experiments,
        },
        "device_settings": self.device_settings,
        "user_agent": self.user_agent,
    }
    with open(self.cookie_fname, "w") as f:
        json.dump(data, f)



import base64
import datetime
import hashlib
import hmac
import json
import logging
import os
import random
import sys
import time
import uuid

import pytz
import requests
import requests.utils
import six.moves.urllib as urllib
from requests_toolbelt import MultipartEncoder
from tqdm import tqdm

from . import config, devices
from .api_login import (
    change_device_simulation,
    generate_all_uuids,
    load_uuid_and_cookie,
    login_flow,
    pre_login_flow,
    reinstall_app_simulation,
    save_uuid_and_cookie,
    set_device,
    sync_device_features,
    sync_launcher,
    sync_user_features,
)
from .api_photo import configure_photo, download_photo, upload_photo
from .api_story import configure_story, download_story, upload_story_photo
from .api_video import configure_video, download_video, upload_video
from .prepare import delete_credentials, get_credentials

try:
    from json.decoder import JSONDecodeError
except ImportError:
    JSONDecodeError = ValueError


PY2 = sys.version_info[0] == 2


class API(object):
    def __init__(
        self,
        device=None,
        base_path="",
        save_logfile=True,
        log_filename=None,
        loglevel_file=logging.INFO,
        loglevel_stream=logging.DEBUG,
    ):
        # Setup device and user_agent
        self.device = device or devices.DEFAULT_DEVICE

        self.cookie_fname = None
        self.base_path = base_path

        self.is_logged_in = False
        self.last_login = None

        self.last_response = None
        self.total_requests = 0

        # Setup logging
        self.logger = logging.getLogger("[instabot_{}]".format(id(self)))

        if not os.path.exists("./config/"):
            os.makedirs("./config/")  # create base_path if not exists

        if not os.path.exists("./log/"):
            os.makedirs("./log/")  # create log folder if not exists

        if save_logfile is True:
            if log_filename is None:
                log_filename = os.path.join(
                    base_path, "./log/instabot_{}.log".format(id(self))
                )

            fh = logging.FileHandler(filename=log_filename)
            fh.setLevel(loglevel_file)
            fh.setFormatter(
                logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
            )

            self.logger.addHandler(fh)

        ch = logging.StreamHandler()
        ch.setLevel(loglevel_stream)
        ch.setFormatter(logging.Formatter("%(asctime)s - %(levelname)s - %(message)s"))

        self.logger.addHandler(ch)
        self.logger.setLevel(logging.DEBUG)

        self.last_json = None

    def set_user(self, username, password, generate_all_uuids=True, set_device=True):
        self.username = username
        self.password = password

        if set_device is True:
            self.set_device()

        if generate_all_uuids is True:
            self.generate_all_uuids()

    def set_contact_point_prefill(self, usage="prefill"):
        data = json.dumps(
            {
                "id": self.uuid,
                "phone_id": self.phone_id,
                "_csrftoken": self.token,
                "usage": usage,
            }
        )
        return self.send_request("accounts/contact_point_prefill/", data, login=True)

    def get_suggested_searches(self, _type="users"):
        return self.send_request(
            "fbsearch/suggested_searches/", self.json_data({"type": _type})
        )

    def read_msisdn_header(self, usage="default"):
        data = json.dumps({"device_id": self.uuid, "mobile_subno_usage": usage})
        return self.send_request(
            "accounts/read_msisdn_header/",
            data,
            login=True,
            headers={"X-DEVICE-ID": self.uuid},
        )

    def log_attribution(self, usage="default"):
        data = json.dumps({"adid": self.advertising_id})
        return self.send_request("attribution/log_attribution/", data, login=True)

    # ====== ALL METHODS IMPORT FROM api_login ====== #
    def sync_device_features(self, login=False):
        return sync_device_features(self, login)

    def sync_launcher(self, login=False):
        return sync_launcher(self, login)

    def sync_user_features(self):
        return sync_user_features(self)

    def pre_login_flow(self):
        return pre_login_flow(self)

    def login_flow(self, just_logged_in=False, app_refresh_interval=1800):
        return login_flow(self, just_logged_in, app_refresh_interval)

    def set_device(self):
        return set_device(self)

    def generate_all_uuids(self):
        return generate_all_uuids(self)

    def reinstall_app_simulation(self):
        return reinstall_app_simulation(self)

    def change_device_simulation(self):
        return change_device_simulation(self)

    def load_uuid_and_cookie(self, load_uuid=True, load_cookie=True):
        return load_uuid_and_cookie(self, load_uuid=load_uuid, load_cookie=load_cookie)

    def save_uuid_and_cookie(self):
        return save_uuid_and_cookie(self)

    def login(
        self,
        username=None,
        password=None,
        force=False,
        proxy=None,
        use_cookie=True,
        use_uuid=True,
        cookie_fname=None,
        ask_for_code=False,
        set_device=True,
        generate_all_uuids=True,
        is_threaded=False,
    ):
        if password is None:
            username, password = get_credentials(username=username)

        set_device = generate_all_uuids = True
        self.set_user(username, password)
        self.session = requests.Session()

        self.proxy = proxy
        self.set_proxy()  # Only happens if `self.proxy`

        self.cookie_fname = cookie_fname
        if self.cookie_fname is None:
            fmt = "{username}_uuid_and_cookie.json"
            cookie_fname = fmt.format(username=username)
            self.cookie_fname = os.path.join(self.base_path, cookie_fname)

        cookie_is_loaded = False
        msg = "Login flow failed, the cookie is broken. Relogin again."

        if use_cookie is True:
            # try:
            if (
                self.load_uuid_and_cookie(load_cookie=use_cookie, load_uuid=use_uuid)
                is True
            ):
                # Check if the token loaded is valid.
                if self.login_flow(False) is True:
                    cookie_is_loaded = True
                    self.save_successful_login()
                else:
                    self.logger.info(msg)
                    set_device = generate_all_uuids = False
                    force = True

        if not cookie_is_loaded and (not self.is_logged_in or force):
            self.session = requests.Session()
            if use_uuid is True:
                if (
                    self.load_uuid_and_cookie(
                        load_cookie=use_cookie, load_uuid=use_uuid
                    )
                    is False
                ):
                    if set_device is True:
                        self.set_device()
                    if generate_all_uuids is True:
                        self.generate_all_uuids()

            self.pre_login_flow()
            data = json.dumps(
                {
                    "phone_id": self.phone_id,
                    "_csrftoken": self.token,
                    "username": self.username,
                    "guid": self.uuid,
                    "device_id": self.device_id,
                    "password": self.password,
                    "login_attempt_count": "0",
                }
            )

            if self.send_request("accounts/login/", data, True):
                self.save_successful_login()
                self.login_flow(True)
                return True

            elif (
                self.last_json.get("error_type", "") == "checkpoint_challenge_required"
            ):
                self.logger.info("Checkpoint challenge required...")
                if ask_for_code is True:
                    solved = self.solve_challenge()
                    if solved:
                        self.save_successful_login()
                        self.login_flow(True)
                        return True
                    else:
                        self.save_failed_login()
                        return False
                else:
                    return False
            elif self.last_json.get("two_factor_required"):
                if self.two_factor_auth():
                    self.save_successful_login()
                    self.login_flow(True)
                    return True
                else:
                    self.save_failed_login()
                    return False
            else:
                self.save_failed_login()
                return False

    def two_factor_auth(self):
        self.logger.info("Two-factor authentication required")
        two_factor_code = input("Enter 2FA verification code: ")
        two_factor_id = self.last_json["two_factor_info"]["two_factor_identifier"]

        login = self.session.post(
            config.API_URL + "accounts/two_factor_login/",
            data={
                "username": self.username,
                "verification_code": two_factor_code,
                "two_factor_identifier": two_factor_id,
                "password": self.password,
                "device_id": self.device_id,
                "ig_sig_key_version": 4,
            },
            allow_redirects=True,
        )

        if login.status_code == 200:
            resp_json = json.loads(login.text)
            if resp_json["status"] != "ok":
                if "message" in resp_json:
                    self.logger.error("Login error: {}".format(resp_json["message"]))
                else:
                    self.logger.error(
                        ('Login error: "{}" status and' " message {}.").format(
                            resp_json["status"], login.text
                        )
                    )
                return False
            return True
        else:
            self.logger.error(
                (
                    "Two-factor authentication request returns "
                    "{} error with message {} !"
                ).format(login.status_code, login.text)
            )
            return False

    def save_successful_login(self):
        self.is_logged_in = True
        self.last_login = time.time()
        self.logger.info("Logged-in successfully as '{}'!".format(self.username))

    def save_failed_login(self):
        self.logger.info("Username or password is incorrect.")
        delete_credentials()

    def solve_challenge(self):
        challenge_url = self.last_json["challenge"]["api_path"][1:]
        try:
            self.send_request(challenge_url, None, login=True, with_signature=False)
        except Exception as e:
            self.logger.error("solve_challenge; {}".format(e))
            return False

        choices = self.get_challenge_choices()
        for choice in choices:
            print(choice)
        code = input("Insert choice: ")

        data = json.dumps({"choice": code})
        try:
            self.send_request(challenge_url, data, login=True)
        except Exception as e:
            self.logger.error(e)
            return False

        print("A code has been sent to the method selected, please check.")
        code = input("Insert code: ").replace(" ", "")

        data = json.dumps({"security_code": code})
        try:
            self.send_request(challenge_url, data, login=True)
        except Exception as e:
            self.logger.error(e)
            return False

        worked = (
            ("logged_in_user" in self.last_json)
            and (self.last_json.get("action", "") == "close")
            and (self.last_json.get("status", "") == "ok")
        )

        if worked:
            return True

        self.logger.error("Not possible to log in. Reset and try again")
        return False

    def get_challenge_choices(self):
        last_json = self.last_json
        choices = []

        if last_json.get("step_name", "") == "select_verify_method":
            choices.append("Checkpoint challenge received")
            if "phone_number" in last_json["step_data"]:
                choices.append("0 - Phone")
            if "email" in last_json["step_data"]:
                choices.append("1 - Email")

        if last_json.get("step_name", "") == "delta_login_review":
            choices.append("Login attempt challenge received")
            choices.append("0 - It was me")
            choices.append("0 - It wasn't me")

        if not choices:
            choices.append(
                '"{}" challenge received'.format(last_json.get("step_name", "Unknown"))
            )
            choices.append("0 - Default")

        return choices

    def logout(self, *args, **kwargs):
        if not self.is_logged_in:
            return True
        data = json.dumps({})
        self.is_logged_in = not self.send_request(
            "accounts/logout/", data, with_signature=False
        )
        return not self.is_logged_in

    def set_proxy(self):
        if getattr(self, "proxy", None):
            parsed = urllib.parse.urlparse(self.proxy)
            scheme = "http://" if not parsed.scheme else ""
            self.session.proxies["http"] = scheme + self.proxy
            self.session.proxies["https"] = scheme + self.proxy

    def send_request(  # noqa: C901 make sleep minutes 0 when we do a request
        self,
        endpoint,
        post=None,
        login=False,
        with_signature=True,
        headers=None,
        extra_sig=None,
        timeout_minutes=None,
    ):
        self.set_proxy()  # Only happens if `self.proxy`
        if not self.is_logged_in and not login:
            msg = "Not logged in!"
            self.logger.critical(msg)
            raise Exception(msg)

        self.session.headers.update(config.REQUEST_HEADERS)
        self.session.headers.update(
            {
                "User-Agent": self.user_agent,
                "X-IG-Connection-Speed": "-1kbps",
                "X-IG-Bandwidth-Speed-KBPS": str(random.randint(7000, 10000)),
                "X-IG-Bandwidth-TotalBytes-B": str(random.randint(500000, 900000)),
                "X-IG-Bandwidth-TotalTime-MS": str(random.randint(50, 150)),
            }
        )
        if headers:
            self.session.headers.update(headers)
        try:
            self.total_requests += 1
            if post is not None:  # POST
                if with_signature:
                    post = self.generate_signature(
                        post
                    )  # Only `send_direct_item` doesn't need a signature
                    if extra_sig is not None and extra_sig != []:
                        post += "&".join(extra_sig)
                response = self.session.post(config.API_URL + endpoint, data=post)
            else:  # GET
                response = self.session.get(config.API_URL + endpoint)
        except Exception as e:
            self.logger.warning(str(e))
            return False

        self.last_response = response
        self.logger.debug(response)
        if response.status_code == 200:
            try:
                self.last_json = json.loads(response.text)
                return True
            except JSONDecodeError:
                return False
        else:
            # print(endpoint, post, response.content)
            if response.status_code != 404 and response.status_code != "404":
                self.logger.error(
                    "Request returns {} error!".format(response.status_code)
                )
            try:
                response_data = json.loads(response.text)
                if response_data.get(
                    "message"
                ) is not None and "feedback_required" in str(
                    response_data.get("message").encode("utf-8")
                ):
                    self.logger.error(
                        "ATTENTION!: `feedback_required`"
                        + str(response_data.get("feedback_message").encode("utf-8"))
                    )
                    try:
                        self.last_response = response
                        self.last_json = json.loads(response.text)
                    except Exception:
                        pass
                    return "feedback_required"
            except ValueError:
                self.logger.error(
                    "Error checking for `feedback_required`, "
                    "response text is not JSON"
                )
                self.logger.info("Full Response: {}".format(str(response)))
                try:
                    self.logger.info("Response Text: {}".format(str(response.text)))
                except Exception:
                    pass

            if response.status_code == 429:
                # if we come to this error, add 5 minutes of sleep everytime we hit the 429 error (aka soft bann) keep increasing untill we are unbanned
                if timeout_minutes is None:
                    timeout_minutes = 0
                timeout_minutes += 1
                self.logger.warning(
                    "That means 'too many requests'. I'll go to sleep "
                    "for {} minutes.".format(timeout_minutes)
                )
                time.sleep(timeout_minutes * 60)
                return self.send_request(
                    endpoint,
                    post,
                    login,
                    with_signature,
                    headers,
                    extra_sig,
                    timeout_minutes,
                )
            elif response.status_code == 400:
                response_data = json.loads(response.text)

                # PERFORM Interactive Two-Factor Authentication
                if response_data.get("two_factor_required"):
                    try:
                        self.last_response = response
                        self.last_json = json.loads(response.text)
                    except Exception:
                        pass
                    return self.two_factor_auth()
                # End of Interactive Two-Factor Authentication
                else:
                    msg = "Instagram's error message: {}"
                    self.logger.info(msg.format(response_data.get("message")))
                    if "error_type" in response_data:
                        msg = "Error type: {}".format(response_data["error_type"])
                    self.logger.info(msg)

            # For debugging
            try:
                self.last_response = response
                self.last_json = json.loads(response.text)
            except Exception:
                pass
            return False

    @property
    def cookie_dict(self):
        return self.session.cookies.get_dict()

    @property
    def token(self):
        return self.cookie_dict["csrftoken"]

    @property
    def user_id(self):
        return self.cookie_dict["ds_user_id"]

    @property
    def rank_token(self):
        return "{}_{}".format(self.user_id, self.uuid)

    @property
    def default_data(self):
        return {"_uuid": self.uuid, "_uid": self.user_id, "_csrftoken": self.token}

    def json_data(self, data=None):
        """Adds the default_data to data and dumps it to a json."""
        if data is None:
            data = {}
        data.update(self.default_data)
        return json.dumps(data)

    def action_data(self, data):
        _data = {"radio_type": "wifi-none", "device_id": self.device_id}
        data.update(_data)
        return data

    def auto_complete_user_list(self):
        return self.send_request("friendships/autocomplete_user_list/")

    def batch_fetch(self):
        data = {
            "scale": 3,
            "version": 1,
            "vc_policy": "default",
            "surfaces_to_triggers": '{"5734":["instagram_feed_prompt"],'
            + '"4715":["instagram_feed_header"],'
            + '"5858":["instagram_feed_tool_tip"]}',  # noqa
            "surfaces_to_queries": (
                '{"5734":"viewer() {eligible_promotions.trigger_context_v2(<tr'
                "igger_context_v2>).ig_parameters(<ig_parameters>).trigger_nam"
                "e(<trigger_name>).surface_nux_id(<surface>).external_gating_p"
                "ermitted_qps(<external_gating_permitted_qps>).supports_client"
                "_filters(true).include_holdouts(true) {edges {client_ttl_seco"
                "nds,log_eligibility_waterfall,is_holdout,priority,time_range "
                "{start,end},node {id,promotion_id,logging_data,max_impression"
                "s,triggers,contextual_filters {clause_type,filters {filter_ty"
                "pe,unknown_action,value {name,required,bool_value,int_value,s"
                "tring_value},extra_datas {name,required,bool_value,int_value,"
                "string_value}},clauses {clause_type,filters {filter_type,unkn"
                "own_action,value {name,required,bool_value,int_value,string_v"
                "alue},extra_datas {name,required,bool_value,int_value,string_"
                "value}},clauses {clause_type,filters {filter_type,unknown_act"
                "ion,value {name,required,bool_value,int_value,string_value},e"
                "xtra_datas {name,required,bool_value,int_value,string_value}}"
                ",clauses {clause_type,filters {filter_type,unknown_action,val"
                "ue {name,required,bool_value,int_value,string_value},extra_da"
                "tas {name,required,bool_value,int_value,string_value}}}}}},is"
                "_uncancelable,template {name,parameters {name,required,bool_v"
                "alue,string_value,color_value,}},creatives {title {text},cont"
                "ent {text},footer {text},social_context {text},social_context"
                "_images,primary_action{title {text},url,limit,dismiss_promoti"
                "on},secondary_action{title {text},url,limit,dismiss_promotion"
                "},dismiss_action{title {text},url,limit,dismiss_promotion},im"
                'age.scale(<scale>) {uri,width,height}}}}}}","4715":"viewer() '
                "{eligible_promotions.trigger_context_v2(<trigger_context_v2>)"
                ".ig_parameters(<ig_parameters>).trigger_name(<trigger_name>)."
                "surface_nux_id(<surface>).external_gating_permitted_qps(<exte"
                "rnal_gating_permitted_qps>).supports_client_filters(true).inc"
                "lude_holdouts(true) {edges {client_ttl_seconds,log_eligibilit"
                "y_waterfall,is_holdout,priority,time_range {start,end},node {"
                "id,promotion_id,logging_data,max_impressions,triggers,context"
                "ual_filters {clause_type,filters {filter_type,unknown_action,"
                "value {name,required,bool_value,int_value,string_value},extra"
                "_datas {name,required,bool_value,int_value,string_value}},cla"
                "uses {clause_type,filters {filter_type,unknown_action,value {"
                "name,required,bool_value,int_value,string_value},extra_datas "
                "{name,required,bool_value,int_value,string_value}},clauses {c"
                "lause_type,filters {filter_type,unknown_action,value {name,re"
                "quired,bool_value,int_value,string_value},extra_datas {name,r"
                "equired,bool_value,int_value,string_value}},clauses {clause_t"
                "ype,filters {filter_type,unknown_action,value {name,required,"
                "bool_value,int_value,string_value},extra_datas {name,required"
                ",bool_value,int_value,string_value}}}}}},is_uncancelable,temp"
                "late {name,parameters {name,required,bool_value,string_value,"
                "color_value,}},creatives {title {text},content {text},footer "
                "{text},social_context {text},social_context_images,primary_ac"
                "tion{title {text},url,limit,dismiss_promotion},secondary_acti"
                "on{title {text},url,limit,dismiss_promotion},dismiss_action{t"
                "itle {text},url,limit,dismiss_promotion},image.scale(<scale>)"
                ' {uri,width,height}}}}}}","5858":"viewer() {eligible_promotio'
                "ns.trigger_context_v2(<trigger_context_v2>).ig_parameters(<ig"
                "_parameters>).trigger_name(<trigger_name>).surface_nux_id(<su"
                "rface>).external_gating_permitted_qps(<external_gating_permit"
                "ted_qps>).supports_client_filters(true).include_holdouts(true"
                ") {edges {client_ttl_seconds,log_eligibility_waterfall,is_hol"
                "dout,priority,time_range {start,end},node {id,promotion_id,lo"
                "gging_data,max_impressions,triggers,contextual_filters {claus"
                "e_type,filters {filter_type,unknown_action,value {name,requir"
                "ed,bool_value,int_value,string_value},extra_datas {name,requi"
                "red,bool_value,int_value,string_value}},clauses {clause_type,"
                "filters {filter_type,unknown_action,value {name,required,bool"
                "_value,int_value,string_value},extra_datas {name,required,boo"
                "l_value,int_value,string_value}},clauses {clause_type,filters"
                " {filter_type,unknown_action,value {name,required,bool_value,"
                "int_value,string_value},extra_datas {name,required,bool_value"
                ",int_value,string_value}},clauses {clause_type,filters {filte"
                "r_type,unknown_action,value {name,required,bool_value,int_val"
                "ue,string_value},extra_datas {name,required,bool_value,int_va"
                "lue,string_value}}}}}},is_uncancelable,template {name,paramet"
                "ers {name,required,bool_value,string_value,color_value,}},cre"
                "atives {title {text},content {text},footer {text},social_cont"
                "ext {text},social_context_images,primary_action{title {text},"
                "url,limit,dismiss_promotion},secondary_action{title {text},ur"
                "l,limit,dismiss_promotion},dismiss_action{title {text},url,li"
                "mit,dismiss_promotion},image.scale(<scale>) {uri,width,height"
                '}}}}}}"}'
            ),  # noqa (Just copied from request)
        }
        data = self.json_data(data)
        return self.send_request("qp/batch_fetch/", data)

    def get_timeline_feed(self, options=[]):
        headers = {"X-Ads-Opt-Out": "0", "X-DEVICE-ID": self.uuid}
        data = {
            "_csrftoken": self.token,
            "_uuid": self.uuid,
            "is_prefetch": 0,
            "phone_id": self.phone_id,
            "device_id": self.uuid,
            "client_session_id": self.client_session_id,
            "battery_level": random.randint(25, 100),
            "is_charging": random.randint(0, 1),
            "will_sound_on": random.randint(0, 1),
            "is_on_screen": True,
            "timezone_offset": datetime.datetime.now(pytz.timezone("CET")).strftime(
                "%z"
            ),
        }

        if "is_pull_to_refresh" in options:
            data["reason"] = "pull_to_refresh"
            data["is_pull_to_refresh"] = "1"
        elif "is_pull_to_refresh" not in options:
            data["reason"] = "cold_start_fetch"
            data["is_pull_to_refresh"] = "0"

        # unseen_posts
        # feed_view_info
        # seen_posts

        if "push_disabled" in options:
            data["push_disabled"] = "true"

        if "recovered_from_crash" in options:
            data["recovered_from_crash"] = "1"

        data = json.dumps(data)
        return self.send_request(
            "feed/timeline/", data, with_signature=False, headers=headers
        )

    def get_megaphone_log(self):
        return self.send_request("megaphone/log/")

    def expose(self):
        data = self.json_data(
            {"id": self.uuid, "experiment": "ig_android_profile_contextual_feed"}
        )
        return self.send_request("qe/expose/", data)

    # ====== PHOTO METHODS ====== #
    def upload_photo(
        self,
        photo,
        caption=None,
        upload_id=None,
        from_video=False,
        force_resize=False,
        options={},
    ):
        """Upload photo to Instagram

        @param photo         Path to photo file (String)
        @param caption       Media description (String)
        @param upload_id     Unique upload_id (String). When None, then
                             generate automatically
        @param from_video    A flag that signals whether the photo is loaded
                             from the video or by itself
                             (Boolean, DEPRECATED: not used)
        @param force_resize  Force photo resize (Boolean)
        @param options       Object with difference options, e.g.
                             configure_timeout, rename (Dict)
                             Designed to reduce the number of function
                             arguments! This is the simplest request object.

        @return Boolean
        """
        return upload_photo(
            self, photo, caption, upload_id, from_video, force_resize, options
        )

    def download_photo(self, media_id, filename, media=False, folder="photos"):
        return download_photo(self, media_id, filename, media, folder)

    def configure_photo(self, upload_id, photo, caption=""):
        return configure_photo(self, upload_id, photo, caption)

    # ====== STORY METHODS ====== #
    def download_story(self, filename, story_url, username):
        return download_story(self, filename, story_url, username)

    def upload_story_photo(self, photo, upload_id=None):
        return upload_story_photo(self, photo, upload_id)

    def configure_story(self, upload_id, photo):
        return configure_story(self, upload_id, photo)

    # ====== VIDEO METHODS ====== #
    def upload_video(
        self, video, caption=None, upload_id=None, thumbnail=None, options={}
    ):
        """Upload video to Instagram

        @param video      Path to video file (String)
        @param caption    Media description (String)
        @param upload_id  Unique upload_id (String). When None, then
                          generate automatically
        @param thumbnail  Path to thumbnail for video (String). When None,
                          then thumbnail is generate automatically
        @param options    Object with difference options, e.g.
                          configure_timeout, rename_thumbnail, rename (Dict)
                          Designed to reduce the number of function arguments!
                          This is the simplest request object.

        @return           Object with state of uploading to
                          Instagram (or False)
        """
        return upload_video(self, video, caption, upload_id, thumbnail, options)

    def download_video(self, media_id, filename, media=False, folder="video"):
        return download_video(self, media_id, filename, media, folder)

    def configure_video(
        self,
        upload_id,
        video,
        thumbnail,
        width,
        height,
        duration,
        caption="",
        options={},
    ):
        """Post Configure Video
        (send caption, thumbnail andmore else to Instagram)

        @param upload_id  Unique upload_id (String). Received
                          from "upload_video"
        @param video      Path to video file (String)
        @param thumbnail  Path to thumbnail for video (String). When None,
                          then thumbnail is generate automatically
        @param width      Width in px (Integer)
        @param height     Height in px (Integer)
        @param duration   Duration in seconds (Integer)
        @param caption    Media description (String)
        @param options    Object with difference options, e.g.
                          configure_timeout, rename_thumbnail, rename (Dict)
                          Designed to reduce the number of function arguments!
                          This is the simplest request object.
        """
        return configure_video(
            self, upload_id, video, thumbnail, width, height, duration, caption, options
        )

    # ====== MEDIA METHODS ====== #
    def edit_media(self, media_id, captionText=""):
        data = self.json_data({"caption_text": captionText})
        url = "media/{media_id}/edit_media/".format(media_id=media_id)
        return self.send_request(url, data)

    def remove_self_tag(self, media_id):
        data = self.json_data()
        url = "media/{media_id}/remove/".format(media_id=media_id)
        return self.send_request(url, data)

    def media_info(self, media_id):
        # data = self.json_data({'media_id': media_id})
        url = "media/{media_id}/info/".format(media_id=media_id)
        return self.send_request(url)

    def archive_media(self, media, undo=False):
        action = "only_me" if not undo else "undo_only_me"
        data = self.json_data({"media_id": media["id"]})
        url = "media/{media_id}/{action}/?media_type={media_type}".format(
            media_id=media["id"], action=action, media_type=media["media_type"]
        )
        return self.send_request(url, data)

    def delete_media(self, media):
        data = self.json_data({"media_id": media.get("id")})
        url = "media/{media_id}/delete/".format(media_id=media.get("id"))
        return self.send_request(url, data)

    def gen_user_breadcrumb(self, size):
        key = "iN4$aGr0m"
        dt = int(time.time() * 1000)

        time_elapsed = random.randint(500, 1500) + size * random.randint(500, 1500)
        text_change_event_count = max(1, size / random.randint(3, 5))

        data = "{size!s} {elapsed!s} {count!s} {dt!s}".format(
            **{
                "size": size,
                "elapsed": time_elapsed,
                "count": text_change_event_count,
                "dt": dt,
            }
        )
        return "{!s}\n{!s}\n".format(
            base64.b64encode(
                hmac.new(
                    key.encode("ascii"), data.encode("ascii"), digestmod=hashlib.sha256
                ).digest()
            ),
            base64.b64encode(data.encode("ascii")),
        )

    def comment(self, media_id, comment_text):
        return self.send_request(
            endpoint="media/{media_id}/comment/".format(media_id=media_id),
            post=self.json_data(
                self.action_data(
                    {
                        "container_module": "comments_v2",
                        "user_breadcrumb": self.gen_user_breadcrumb(len(comment_text)),
                        "idempotence_token": self.generate_UUID(True),
                        "comment_text": comment_text,
                    }
                )
            ),
        )

    def reply_to_comment(self, media_id, comment_text, parent_comment_id):
        data = self.json_data(
            {"comment_text": comment_text, "replied_to_comment_id": parent_comment_id}
        )
        url = "media/{media_id}/comment/".format(media_id=media_id)
        return self.send_request(url, data)

    def delete_comment(self, media_id, comment_id):
        data = self.json_data()
        url = "media/{media_id}/comment/{comment_id}/delete/"
        url = url.format(media_id=media_id, comment_id=comment_id)
        return self.send_request(url, data)

    def get_comment_likers(self, comment_id):
        url = "media/{comment_id}/comment_likers/?".format(comment_id=comment_id)
        return self.send_request(url)

    def get_media_likers(self, media_id):
        url = "media/{media_id}/likers/?".format(media_id=media_id)
        return self.send_request(url)

    def like_comment(self, comment_id):
        data = self.json_data(
            {
                "is_carousel_bumped_post": "false",
                "container_module": "comments_v2",
                "feed_position": "0",
            }
        )
        url = "media/{comment_id}/comment_like/".format(comment_id=comment_id)
        return self.send_request(url, data)

    def unlike_comment(self, comment_id):
        data = self.json_data(
            {
                "is_carousel_bumped_post": "false",
                "container_module": "comments_v2",
                "feed_position": "0",
            }
        )
        url = "media/{comment_id}/comment_unlike/".format(comment_id=comment_id)
        return self.send_request(url, data)

    # From profile => "is_carousel_bumped_post":"false",
    # "container_module":"feed_contextual_profile", "feed_position":"0" # noqa
    # From home/feed => "inventory_source":"media_or_ad",
    # "is_carousel_bumped_post":"false", "container_module":"feed_timeline",
    # "feed_position":"0" # noqa
    def like(
        self,
        media_id,
        double_tap=None,
        container_module="feed_short_url",
        feed_position=0,
        username=None,
        user_id=None,
        hashtag_name=None,
        hashtag_id=None,
        entity_page_name=None,
        entity_page_id=None,
    ):

        data = self.action_data(
            {
                "media_id": media_id,
                "container_module": container_module,
                "feed_position": str(feed_position),
                "is_carousel_bumped_post": "false",
            }
        )
        if container_module == "feed_timeline":
            data.update({"inventory_source": "media_or_ad"})
        if username:
            data.update({"username": username, "user_id": user_id})
        if hashtag_name:
            data.update({"hashtag_name": hashtag_name, "hashtag_id": hashtag_id})
        if entity_page_name:
            data.update(
                {"entity_page_name": entity_page_name, "entity_page_id": entity_page_id}
            )
        if double_tap is None:
            double_tap = random.randint(0, 1)
        json_data = self.json_data(data)
        # TODO: comment out debug log out when done
        self.logger.debug("post data: {}".format(json_data))
        return self.send_request(
            endpoint="media/{media_id}/like/".format(media_id=media_id),
            post=json_data,
            extra_sig=["d={}".format(double_tap)],
        )

    def unlike(self, media_id):
        data = self.json_data(
            {
                "media_id": media_id,
                "radio_type": "wifi-none",
                "is_carousel_bumped_post": "false",
                "container_module": "photo_view_other",
                "feed_position": "0",
            }
        )
        url = "media/{media_id}/unlike/".format(media_id=media_id)
        return self.send_request(url, data)

    def get_media_comments(self, media_id, max_id=""):
        url = "media/{media_id}/comments/".format(media_id=media_id)
        if max_id:
            url += "?max_id={max_id}".format(max_id=max_id)
        return self.send_request(url)

    def explore(self, is_prefetch=False):
        data = {
            "is_prefetch": is_prefetch,
            "is_from_promote": False,
            "timezone_offset": datetime.datetime.now(pytz.timezone("CET")).strftime(
                "%z"
            ),
            "session_id": self.client_session_id,
            "supported_capabilities_new": config.SUPPORTED_CAPABILITIES,
        }
        if is_prefetch:
            data["max_id"] = 0
            data["module"] = "explore_popular"
        data = json.dumps(data)
        return self.send_request("discover/explore/", data)

    def get_username_info(self, user_id):
        url = "users/{user_id}/info/".format(user_id=user_id)
        return self.send_request(url)

    def get_self_username_info(self):
        return self.get_username_info(self.user_id)

    def get_recent_activity(self):
        return self.send_request("news/inbox")

    def get_following_recent_activity(self):
        return self.send_request("news")

    def get_user_tags(self, user_id):
        url = (
            "usertags/{user_id}/feed/?rank_token=" "{rank_token}&ranked_content=true&"
        ).format(user_id=user_id, rank_token=self.rank_token)
        return self.send_request(url)

    def get_self_user_tags(self):
        return self.get_user_tags(self.user_id)

    def get_geo_media(self, user_id):
        url = "maps/user/{user_id}/".format(user_id=user_id)
        return self.send_request(url)

    def get_self_geo_media(self):
        return self.get_geo_media(self.user_id)

    def sync_from_adress_book(self, contacts):
        url = "address_book/link/?include=extra_display_name,thumbnails"
        return self.send_request(url, "contacts=" + json.dumps(contacts))

    # ====== FEED METHODS ====== #
    def tag_feed(self, tag):
        url = "feed/tag/{tag}/?rank_token={rank_token}&ranked_content=true&"
        return self.send_request(url.format(tag=tag, rank_token=self.rank_token))

    def get_timeline(self):
        url = "feed/timeline/?rank_token={rank_token}&ranked_content=true&"
        return self.send_request(url.format(rank_token=self.rank_token))

    def get_archive_feed(self):
        url = "feed/only_me_feed/?rank_token={rank_token}&ranked_content=true&"
        return self.send_request(url.format(rank_token=self.rank_token))

    def get_user_feed(self, user_id, max_id="", min_timestamp=None):
        url = (
            "feed/user/{user_id}/?max_id={max_id}&min_timestamp="
            "{min_timestamp}&rank_token={rank_token}&ranked_content=true"
            # noqa
        ).format(
            user_id=user_id,
            max_id=max_id,
            min_timestamp=min_timestamp,
            rank_token=self.rank_token,
        )
        return self.send_request(url)

    def get_self_user_feed(self, max_id="", min_timestamp=None):
        return self.get_user_feed(self.user_id, max_id, min_timestamp)

    def get_hashtag_feed(self, hashtag, max_id=""):
        url = (
            "feed/tag/{hashtag}/?max_id={max_id}"
            "&rank_token={rank_token}&ranked_content=true&"
        ).format(hashtag=hashtag, max_id=max_id, rank_token=self.rank_token)
        return self.send_request(url)

    def get_location_feed(self, location_id, max_id=""):
        url = (
            "feed/location/{location_id}/?max_id={max_id}"
            "&rank_token={rank_token}&ranked_content=true&"
        ).format(location_id=location_id, max_id=max_id, rank_token=self.rank_token)
        return self.send_request(url)

    def get_popular_feed(self):
        url = (
            "feed/popular/?people_teaser_supported=1"
            "&rank_token={rank_token}&ranked_content=true&"
        )
        return self.send_request(url.format(rank_token=self.rank_token))

    def get_liked_media(self, max_id=""):
        url = "feed/liked/?max_id={max_id}".format(max_id=max_id)
        return self.send_request(url)

    # ====== FRIENDSHIPS METHODS ====== #
    def get_user_followings(self, user_id, max_id=""):
        url = (
            "friendships/{user_id}/following/?max_id={max_id}"
            "&ig_sig_key_version={sig_key}&rank_token={rank_token}"
        ).format(
            user_id=user_id,
            max_id=max_id,
            sig_key=config.SIG_KEY_VERSION,
            rank_token=self.rank_token,
        )
        return self.send_request(url)

    def get_self_users_following(self):
        return self.get_user_followings(self.user_id)

    def get_user_followers(self, user_id, max_id=""):
        url = "friendships/{user_id}/followers/?rank_token={rank_token}"
        url = url.format(user_id=user_id, rank_token=self.rank_token)
        if max_id:
            url += "&max_id={max_id}".format(max_id=max_id)
        return self.send_request(url)

    def get_self_user_followers(self):
        return self.followers

    def follow(self, user_id):
        data = self.json_data(self.action_data({"user_id": user_id}))
        self.logger.debug("post data: {}".format(data))
        url = "friendships/create/{user_id}/".format(user_id=user_id)
        return self.send_request(url, data)

    def unfollow(self, user_id):
        data = self.json_data({"user_id": user_id, "radio_type": "wifi-none"})
        url = "friendships/destroy/{user_id}/".format(user_id=user_id)
        return self.send_request(url, data)

    def remove_follower(self, user_id):
        data = self.json_data({"user_id": user_id})
        url = "friendships/remove_follower/{user_id}/".format(user_id=user_id)
        return self.send_request(url, data)

    def block(self, user_id):
        data = self.json_data({"user_id": user_id})
        url = "friendships/block/{user_id}/".format(user_id=user_id)
        return self.send_request(url, data)

    def unblock(self, user_id):
        data = self.json_data({"user_id": user_id})
        url = "friendships/unblock/{user_id}/".format(user_id=user_id)
        return self.send_request(url, data)

    def user_friendship(self, user_id):
        data = self.json_data({"user_id": user_id})
        url = "friendships/show/{user_id}/".format(user_id=user_id)
        return self.send_request(url, data)

    def mute_user(self, user, mute_story=False, mute_posts=False):
        data_dict = {}
        if mute_posts:
            data_dict["target_posts_author_id"] = user
        if mute_story:
            data_dict["target_reel_author_id"] = user
        data = self.json_data(data_dict)
        url = "friendships/mute_posts_or_story_from_follow/"
        return self.send_request(url, data)

    def get_muted_friends(self, muted_content):
        # ToDo update endpoints for posts
        if muted_content == "stories":
            url = "friendships/muted_reels"
        elif muted_content == "posts":
            raise NotImplementedError(
                "API does not support getting friends "
                "with muted {}".format(muted_content)
            )
        else:
            raise NotImplementedError(
                "API does not support getting friends"
                " with muted {}".format(muted_content)
            )

        return self.send_request(url)

    def unmute_user(self, user, unmute_posts=False, unmute_stories=False):
        data_dict = {}
        if unmute_posts:
            data_dict["target_posts_author_id"] = user
        if unmute_stories:
            data_dict["target_reel_author_id"] = user
        data = self.json_data(data_dict)
        url = "friendships/unmute_posts_or_story_from_follow/"
        return self.send_request(url, data)

    def get_pending_friendships(self):
        """Get pending follow requests"""
        url = "friendships/pending/"
        return self.send_request(url)

    def approve_pending_friendship(self, user_id):
        data = self.json_data(
            {
                "_uuid": self.uuid,
                "_uid": self.user_id,
                "user_id": user_id,
                "_csrftoken": self.token,
            }
        )
        url = "friendships/approve/{}/".format(user_id)
        return self.send_request(url, post=data)

    def reject_pending_friendship(self, user_id):
        data = self.json_data(
            {
                "_uuid": self.uuid,
                "_uid": self.user_id,
                "user_id": user_id,
                "_csrftoken": self.token,
            }
        )
        url = "friendships/ignore/{}/".format(user_id)
        return self.send_request(url, post=data)

    def get_direct_share(self):
        return self.send_request("direct_share/inbox/?")

    @staticmethod
    def _prepare_recipients(users, thread_id=None, use_quotes=False):
        if not isinstance(users, list):
            print("Users must be an list")
            return False
        result = {"users": "[[{}]]".format(",".join(users))}
        if thread_id:
            template = '["{}"]' if use_quotes else "[{}]"
            result["thread"] = template.format(thread_id)
        return result

    @staticmethod
    def generate_signature(data):
        body = (
            hmac.new(
                config.IG_SIG_KEY.encode("utf-8"), data.encode("utf-8"), hashlib.sha256
            ).hexdigest()
            + "."
            + urllib.parse.quote(data)
        )
        signature = "ig_sig_key_version={sig_key}&signed_body={body}"
        return signature.format(sig_key=config.SIG_KEY_VERSION, body=body)

    @staticmethod
    def generate_device_id(seed):
        volatile_seed = "12345"
        m = hashlib.md5()
        m.update(seed.encode("utf-8") + volatile_seed.encode("utf-8"))
        return "android-" + m.hexdigest()[:16]

    @staticmethod
    def get_seed(*args):
        m = hashlib.md5()
        m.update(b"".join([arg.encode("utf-8") for arg in args]))
        return m.hexdigest()

    @staticmethod
    def generate_UUID(uuid_type):
        generated_uuid = str(uuid.uuid4())
        if uuid_type:
            return generated_uuid
        else:
            return generated_uuid.replace("-", "")

    def get_total_followers_or_followings(  # noqa: C901
        self,
        user_id,
        amount=None,
        which="followers",
        filter_private=False,
        filter_business=False,
        filter_verified=False,
        usernames=False,
        to_file=None,
        overwrite=False,
    ):
        from io import StringIO

        if which == "followers":
            key = "follower_count"
            get = self.get_user_followers
        elif which == "followings":
            key = "following_count"
            get = self.get_user_followings

        sleep_track = 0
        result = []
        next_max_id = ""
        self.get_username_info(user_id)
        username_info = self.last_json
        if "user" in username_info:
            total = amount or username_info["user"][key]

            if total > 200000:
                print(
                    "Consider temporarily saving the result of this big "
                    "operation. This will take a while.\n"
                )
        else:
            return False
        if filter_business:
            print(
                "--> You are going to filter business accounts. "
                "This will take time! <--"
            )
        if to_file is not None:
            if os.path.isfile(to_file):
                if not overwrite:
                    print("File `{}` already exists. Not overwriting.".format(to_file))
                    return False
                else:
                    print("Overwriting file `{}`".format(to_file))
            with open(to_file, "w"):
                pass
        desc = "Getting {} of {}".format(which, user_id)
        with tqdm(total=total, desc=desc, leave=True) as pbar:
            while True:
                get(user_id, next_max_id)
                last_json = self.last_json
                try:
                    with open(to_file, "a") if to_file is not None else StringIO() as f:
                        for item in last_json["users"]:
                            if filter_private and item["is_private"]:
                                continue
                            if filter_business:
                                time.sleep(2 * random.random())
                                self.get_username_info(item["pk"])
                                item_info = self.last_json
                                if item_info["user"]["is_business"]:
                                    continue
                            if filter_verified and item["is_verified"]:
                                continue
                            if to_file is not None:
                                if usernames:
                                    f.write("{}\n".format(item["username"]))
                                else:
                                    f.write("{}\n".format(item["pk"]))
                            result.append(item)
                            pbar.update(1)
                            sleep_track += 1
                            if sleep_track >= 20000:
                                sleep_time = random.uniform(120, 180)
                                msg = (
                                    "\nWaiting {:.2f} min. " "due to too many requests."
                                ).format(sleep_time / 60)
                                print(msg)
                                time.sleep(sleep_time)
                                sleep_track = 0
                    if not last_json["users"] or len(result) >= total:
                        return result[:total]
                except Exception as e:
                    print("ERROR: {}".format(e))
                    return result[:total]

                if last_json["big_list"] is False:
                    return result[:total]

                next_max_id = last_json.get("next_max_id", "")

    def get_total_followers(self, user_id, amount=None):
        return self.get_total_followers_or_followings(user_id, amount, "followers")

    def get_total_followings(self, user_id, amount=None):
        return self.get_total_followers_or_followings(user_id, amount, "followings")

    def get_total_user_feed(self, user_id, min_timestamp=None):
        return self.get_last_user_feed(
            user_id, amount=float("inf"), min_timestamp=min_timestamp
        )

    def get_last_user_feed(self, user_id, amount, min_timestamp=None):
        user_feed = []
        next_max_id = ""
        while True:
            if len(user_feed) >= float(amount):
                # one request returns max 13 items
                return user_feed[:amount]
            self.get_user_feed(user_id, next_max_id, min_timestamp)
            last_json = self.last_json
            if "items" not in last_json:
                return user_feed
            user_feed += last_json["items"]
            if not last_json.get("more_available"):
                return user_feed
            next_max_id = last_json.get("next_max_id", "")

    def get_total_hashtag_feed(self, hashtag_str, amount=100):
        hashtag_feed = []
        next_max_id = ""

        with tqdm(total=amount, desc="Getting hashtag media.", leave=False) as pbar:
            while True:
                self.get_hashtag_feed(hashtag_str, next_max_id)
                last_json = self.last_json
                if "items" not in last_json:
                    return hashtag_feed[:amount]
                items = last_json["items"]
                try:
                    pbar.update(len(items))
                    hashtag_feed += items
                    if not items or len(hashtag_feed) >= amount:
                        return hashtag_feed[:amount]
                except Exception:
                    return hashtag_feed[:amount]
                next_max_id = last_json.get("next_max_id", "")

    def get_total_self_user_feed(self, min_timestamp=None):
        return self.get_total_user_feed(self.user_id, min_timestamp)

    def get_total_self_followers(self):
        return self.get_total_followers(self.user_id)

    def get_total_self_followings(self):
        return self.get_total_followings(self.user_id)

    def get_total_liked_media(self, scan_rate=1):
        next_id = ""
        liked_items = []
        for _ in range(scan_rate):
            self.get_liked_media(next_id)
            last_json = self.last_json
            next_id = last_json.get("next_max_id", "")
            liked_items += last_json["items"]
        return liked_items

    # ====== ACCOUNT / PERSONAL INFO METHODS ====== #
    def change_password(self, new_password):
        data = self.json_data(
            {
                "old_password": self.password,
                "new_password1": new_password,
                "new_password2": new_password,
            }
        )
        return self.send_request("accounts/change_password/", data)

    def remove_profile_picture(self):
        data = self.json_data()
        return self.send_request("accounts/remove_profile_picture/", data)

    def set_private_account(self):
        data = self.json_data()
        return self.send_request("accounts/set_private/", data)

    def set_public_account(self):
        data = self.json_data()
        return self.send_request("accounts/set_public/", data)

    def set_name_and_phone(self, name="", phone=""):
        return self.send_request(
            "accounts/set_phone_and_name/",
            self.json_data({"first_name": name, "phone_number": phone}),
        )

    def get_profile_data(self):
        data = self.json_data()
        return self.send_request("accounts/current_user/?edit=true", data)

    def edit_profile(self, url, phone, first_name, biography, email, gender):
        data = self.json_data(
            {
                "external_url": url,
                "phone_number": phone,
                "username": self.username,
                "full_name": first_name,
                "biography": biography,
                "email": email,
                "gender": gender,
            }
        )
        return self.send_request("accounts/edit_profile/", data)

    def fb_user_search(self, query):
        url = (
            "fbsearch/topsearch/?context=blended&query={query}"
            "&rank_token={rank_token}"
        )
        return self.send_request(url.format(query=query, rank_token=self.rank_token))

    def search_users(self, query):
        url = (
            "users/search/?ig_sig_key_version={sig_key}"
            "&is_typeahead=true&query={query}&rank_token={rank_token}"
        )
        return self.send_request(
            url.format(
                sig_key=config.SIG_KEY_VERSION, query=query, rank_token=self.rank_token
            )
        )

    def search_username(self, username):
        url = "users/{username}/usernameinfo/".format(username=username)
        return self.send_request(url)

    def search_tags(self, query):
        url = "tags/search/?is_typeahead=true&q={query}" "&rank_token={rank_token}"
        return self.send_request(url.format(query=query, rank_token=self.rank_token))

    def search_location(self, query="", lat=None, lng=None):
        url = (
            "fbsearch/places/?rank_token={rank_token}"
            "&query={query}&lat={lat}&lng={lng}"
        )
        url = url.format(rank_token=self.rank_token, query=query, lat=lat, lng=lng)
        return self.send_request(url)

    def get_user_reel(self, user_id):
        url = "feed/user/{}/reel_media/".format(user_id)
        return self.send_request(url)

    def get_reels_tray_feed(
        self, reason="pull_to_refresh"
    ):  # reason can be = cold_start, pull_to_refresh
        data = {
            "supported_capabilities_new": config.SUPPORTED_CAPABILITIES,
            "reason": reason,
            "_csrftoken": self.token,
            "_uuid": self.uuid,
        }
        data = json.dumps(data)
        return self.send_request("feed/reels_tray/", data)

    def get_users_reel(self, user_ids):
        """
            Input: user_ids - a list of user_id
            Output: dictionary: user_id - stories data.
            Basically, for each user output the same as after
            self.get_user_reel
        """
        url = "feed/reels_media/"
        res = self.send_request(
            url, post=self.json_data({"user_ids": [str(x) for x in user_ids]})
        )
        if res:
            return self.last_json["reels"] if "reels" in self.last_json else []
        return []

    def see_reels(self, reels):
        """
            Input - the list of reels jsons
            They can be aquired by using get_users_reel()
            or get_user_reel() methods
        """
        if not isinstance(reels, list):
            # In case of only one reel as input
            reels = [reels]

        story_seen = {}
        now = int(time.time())
        for i, story in enumerate(
            sorted(reels, key=lambda m: m["taken_at"], reverse=True)
        ):
            story_seen_at = now - min(
                i + 1 + random.randint(0, 2), max(0, now - story["taken_at"])
            )
            story_seen["{!s}_{!s}".format(story["id"], story["user"]["pk"])] = [
                "{!s}_{!s}".format(story["taken_at"], story_seen_at)
            ]

        data = self.json_data(
            {
                "reels": story_seen,
                "_csrftoken": self.token,
                "_uuid": self.uuid,
                "_uid": self.user_id,
            }
        )
        data = self.generate_signature(data)
        return self.session.post(
            "https://i.instagram.com/api/v2/" + "media/seen/", data=data
        ).ok

    def get_user_stories(self, user_id):
        url = "feed/user/{}/story/".format(user_id)
        return self.send_request(url)

    def get_self_story_viewers(self, story_id):
        url = ("media/{}/list_reel_media_viewer/?supported_capabilities_new={}").format(
            story_id, config.SUPPORTED_CAPABILITIES
        )
        return self.send_request(url)

    def get_tv_suggestions(self):
        url = "igtv/tv_guide/"
        return self.send_request(url)

    def get_hashtag_stories(self, hashtag):
        url = "tags/{}/story/".format(hashtag)
        return self.send_request(url)

    def follow_hashtag(self, hashtag):
        data = self.json_data({})
        url = "tags/follow/{}/".format(hashtag)
        return self.send_request(url, data)

    def unfollow_hashtag(self, hashtag):
        data = self.json_data({})
        url = "tags/unfollow/{}/".format(hashtag)
        return self.send_request(url, data)

    def get_tags_followed_by_user(self, user_id):
        url = "users/{}/following_tags_info/".format(user_id)
        return self.send_request(url)

    def get_hashtag_sections(self, hashtag):
        data = self.json_data(
            {
                "supported_tabs": "['top','recent','places']",
                "include_persistent": "true",
            }
        )
        url = "tags/{}/sections/".format(hashtag)
        return self.send_request(url, data)

    def get_media_insight(self, media_id):
        url = ("insights/media_organic_insights/{}/?ig_sig_key_version={}").format(
            media_id, config.IG_SIG_KEY
        )
        return self.send_request(url)

    def get_self_insight(self):
        # TODO:
        url = (
            "insights/account_organic_insights/?"
            "show_promotions_in_landing_page=true&first={}"
        ).format()
        return self.send_request(url)

    # From profile => "module_name":"feed_contextual_profile"
    # From home/feed => "module_name":"feed_timeline"
    def save_media(self, media_id, module_name="feed_timeline"):
        return self.send_request(
            endpoint="media/{media_id}/save/".format(media_id=media_id),
            post=self.json_data(self.action_data({"module_name": module_name})),
        )

    def unsave_media(self, media_id):
        data = self.json_data()
        url = "media/{}/unsave/".format(media_id)
        return self.send_request(url, data)

    def get_saved_medias(self):
        url = "feed/saved/"
        return self.send_request(url)

    def get_loom_fetch_config(self):
        return self.send_request("loom/fetch_config/")

    def get_profile_notice(self):
        return self.send_request("users/profile_notice/")

    # ====== DIRECT METHODS ====== #
    def get_inbox_v2(self):
        data = json.dumps({"persistentBadging": True, "use_unified_inbox": True})
        return self.send_request("direct_v2/inbox/", data)

    def get_presence(self):
        return self.send_request("direct_v2/get_presence/")

    def get_thread(self, thread_id, cursor_id=None):
        data = {"use_unified_inbox": "true"}
        if cursor_id is not None:
            data["cursor"] = cursor_id
        return self.send_request(
            "direct_v2/threads/{}/".format(thread_id), json.dumps(data)
        )

    def get_ranked_recipients(self, mode, show_threads, query=None):
        data = {
            "mode": mode,
            "show_threads": "false" if show_threads is False else "true",
            "use_unified_inbox": "true",
        }
        if query is not None:
            data["query"] = query
        return self.send_request("direct_v2/ranked_recipients/", json.dumps(data))

    def send_direct_item(self, item_type, users, **options):
        data = {"client_context": self.generate_UUID(True), "action": "send_item"}
        headers = {}
        recipients = self._prepare_recipients(
            users, options.get("thread"), use_quotes=False
        )
        if not recipients:
            return False
        data["recipient_users"] = recipients.get("users")
        if recipients.get("thread"):
            data["thread_ids"] = recipients.get("thread")
        data.update(self.default_data)

        url = "direct_v2/threads/broadcast/{}/".format(item_type)
        text = options.get("text", "")
        if item_type == "link":
            data["link_text"] = text
            data["link_urls"] = json.dumps(options.get("urls"))
        elif item_type == "text":
            data["text"] = text
        elif item_type == "media_share":
            data["text"] = text
            data["media_type"] = options.get("media_type", "photo")
            data["media_id"] = options.get("media_id", "")
        elif item_type == "hashtag":
            data["text"] = text
            data["hashtag"] = options.get("hashtag", "")
        elif item_type == "profile":
            data["text"] = text
            data["profile_user_id"] = options.get("profile_user_id")
        elif item_type == "photo":
            url = "direct_v2/threads/broadcast/upload_photo/"
            filepath = options["filepath"]
            upload_id = str(int(time.time() * 1000))
            with open(filepath, "rb") as f:
                photo = f.read()

            data["photo"] = (
                "direct_temp_photo_%s.jpg" % upload_id,
                photo,
                "application/octet-stream",
                {"Content-Transfer-Encoding": "binary"},
            )

            m = MultipartEncoder(data, boundary=self.uuid)
            data = m.to_string()
            headers.update({"Content-type": m.content_type})

        return self.send_request(url, data, with_signature=False, headers=headers)

    def get_pending_inbox(self):
        url = (
            "direct_v2/pending_inbox/?persistentBadging=true" "&use_unified_inbox=true"
        )
        return self.send_request(url)

    # ACCEPT button in pending request
    def approve_pending_thread(self, thread_id):
        data = self.json_data({"_uuid": self.uuid, "_csrftoken": self.token})
        url = "direct_v2/threads/{}/approve/".format(thread_id)
        return self.send_request(url, post=data)

    # DELETE button in pending request
    def hide_pending_thread(self, thread_id):
        data = self.json_data({"_uuid": self.uuid, "_csrftoken": self.token})
        url = "direct_v2/threads/{}/hide/".format(thread_id)
        return self.send_request(url, post=data)

    # BLOCK button in pending request
    def decline_pending_thread(self, thread_id):
        data = self.json_data({"_uuid": self.uuid, "_csrftoken": self.token})
        url = "direct_v2/threads/{}/decline/".format(thread_id)
        return self.send_request(url, post=data)

    def open_instagram_link(self, link):
        return self.send_request(
            "oembed/?url={}".format(urllib.parse.quote(link, safe=""))
        )
